# tests.test_base.py
# Assertions for the base classes and abstract hierarchy.
#
# Author:   Rebecca Bilbro
# Author:   Benjamin Bengfort
# Author:   Neal Humphrey
# Created:  Sat Oct 08 18:34:30 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_base.py [83131ef] benjamin@bengfort.com $

"""
Assertions for the base classes and abstract hierarchy.
"""

##########################################################################
## Imports
##########################################################################

import pytest
import matplotlib.pyplot as plt

from yellowbrick.base import *
from yellowbrick.base import VisualizerGrid
from yellowbrick.datasets import load_occupancy
from yellowbrick.exceptions import YellowbrickWarning
from yellowbrick.exceptions import YellowbrickValueError

from unittest.mock import patch
from unittest.mock import MagicMock
from tests.rand import RandomVisualizer
from tests.base import IS_WINDOWS_OR_CONDA, VisualTestCase

from sklearn.svm import LinearSVC
from sklearn.datasets import make_classification

##########################################################################
## Base Cases
##########################################################################


class TestBaseClasses(VisualTestCase):
    """
    Tests for the high-level API of Yellowbrick and base classes
    """

    def test_visualizer_ax_property(self):
        """
        Test the ax property on the Visualizer
        """
        viz = Visualizer()
        assert viz._ax is None
        assert viz.ax is not None

        viz.ax = "foo"
        assert viz._ax == "foo"
        assert viz.ax == "foo"

    def test_visualizer_fig_property(self):
        """
        Test the fig property on the Visualizer
        """
        viz = Visualizer()
        assert viz._fig is None
        assert viz.fig is not None

        viz.fig = "foo"
        assert viz._fig == "foo"
        assert viz.fig == "foo"

    def test_size_property(self):
        """
        Test the size property on the base Visualizer
        """
        fig = plt.figure(figsize=(1, 2))
        viz = Visualizer()

        assert viz._size is None
        assert viz.size is not None

        fsize = fig.get_size_inches() * fig.get_dpi()
        assert all(viz.size) == all(fsize)

        viz.size = (1080, 720)
        assert viz._size == (1080, 720)
        assert viz.size == (1080, 720)

        fsize = fig.get_size_inches() * fig.get_dpi()
        assert all(viz.size) == all(fsize)

    def test_visualizer_fit_returns_self(self):
        """
        Assert that all visualizers return self
        """
        viz = Visualizer()
        assert viz.fit([]) is viz

    def test_draw_interface(self):
        """
        Assert that draw cannot be called at the base level
        """
        with pytest.raises(NotImplementedError):
            viz = Visualizer()
            viz.draw()

    def test_finalize_interface(self):
        """
        Assert finalize returns the finalized axes
        """
        viz = Visualizer()
        assert viz.finalize() is viz.ax

    @patch("yellowbrick.base.plt")
    def test_show_interface(self, mock_plt):
        """
        Test show calls plt.show and other figure finalization correctly
        """

        class CustomVisualizer(Visualizer):
            pass

        _, ax = plt.subplots()
        viz = CustomVisualizer(ax=ax)
        viz.finalize = MagicMock()
        assert viz.show() is ax

        viz.finalize.assert_called_once_with()
        mock_plt.show.assert_called_once_with()
        mock_plt.savefig.assert_not_called()

    @patch("yellowbrick.base.plt")
    def test_show_savefig_interface(self, mock_plt):
        """
        Test show calls plt.savefig and other figure finalization correctly
        """

        class CustomVisualizer(Visualizer):
            pass

        _, ax = plt.subplots()
        viz = CustomVisualizer(ax=ax)
        viz.finalize = MagicMock()
        assert viz.show(outpath="test.png") is ax

        viz.finalize.assert_called_once_with()
        mock_plt.show.assert_not_called()
        mock_plt.savefig.assert_called_once_with("test.png")

    @patch("yellowbrick.base.plt")
    def test_show_warns(self, mock_plt):
        """
        Test show issues a warning when no axes has been modified
        """

        class CustomVisualizer(Visualizer):
            pass

        with pytest.warns(YellowbrickWarning):
            viz = CustomVisualizer()
            assert viz.show() is not None

    def test_poof_deprecated(self):
        """
        Test that poof issues a deprecation warning
        """

        class CustomVisualizer(Visualizer):
            pass

        viz = CustomVisualizer()
        viz.show = MagicMock()

        with pytest.warns(DeprecationWarning, match="please use show"):
            viz.poof()

        viz.show.assert_called_once()


##########################################################################
## ModelVisualizer Cases
##########################################################################

class MockModelVisualizer(ModelVisualizer):

    def __init__(
        self,
        estimator,
        ax=None,
        fig=None,
        is_fitted="auto",
        param1='foo',
        param2='bar',
        **kwargs
    ):
        super().__init__(estimator, ax=ax, fig=fig, is_fitted=is_fitted, **kwargs)
        self.param1 = param1
        self.param2 = param2


class TestModelVisualizer(VisualTestCase):
    """
    Tests for the ModelVisualizer and wrapped estimator.
    """

    def test_get_params(self):
        """
        Test get_params from visualizer and wrapped estimator
        """
        model = LinearSVC()
        oz = MockModelVisualizer(model)
        params = oz.get_params()

        for key, val in model.get_params().items():
            assert key in params
            assert params[key] == val

        for key in ('ax', 'fig', 'is_fitted', 'param1', 'param2'):
            assert key in params

    def test_set_params(self):
        """
        Test set_params from visualizer to wrapped estimator
        """
        model = LinearSVC()
        oz = MockModelVisualizer(model)

        orig = oz.get_params()
        changes = {
            "param1": 'water',
            "is_fitted": False,
            "estimator__C": 8.8,
            "intercept_scaling": 4.2,
            "dual": False,
        }

        oz.set_params(**changes)
        params = oz.get_params()

        for key, val in params.items():
            if key in changes:
                assert val == changes[key]
            elif key == "C":
                assert val == changes["estimator__C"]
            else:
                assert val == orig[key]


##########################################################################
## ScoreVisualizer Cases
##########################################################################

class MockVisualizer(ScoreVisualizer):
    """
    Mock for a downstream score visualizer
    """

    def fit(self, X, y):
        super(MockVisualizer, self).fit(X, y)


class TestScoreVisualizer(VisualTestCase):
    """
    Tests for the ScoreVisualizer
    """

    def test_with_fitted(self):
        """
        Test that visualizer properly handles an already-fitted model
        """
        X, y = load_occupancy(return_dataset=True).to_numpy()

        model = LinearSVC().fit(X, y)
        classes = ["unoccupied", "occupied"]

        with patch.object(model, "fit") as mockfit:
            oz = MockVisualizer(model, classes=classes)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with patch.object(model, "fit") as mockfit:
            oz = MockVisualizer(model, classes=classes, is_fitted=True)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with patch.object(model, "fit") as mockfit:
            oz = MockVisualizer(model, classes=classes, is_fitted=False)
            oz.fit(X, y)
            mockfit.assert_called_once_with(X, y)


##########################################################################
## Visual Grid Cases
##########################################################################


@pytest.mark.filterwarnings("ignore:Matplotlib is currently using agg")
class TestVisualizerGrid(VisualTestCase):
    """
    Tests for the VisualizerGrid layout class
    """

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_draw_visualizer_grid(self):
        """
        Draw a 4 visualizers grid with default options
        """
        visualizers = [RandomVisualizer(random_state=(1 + x) ** 2) for x in range(4)]

        X, y = make_classification(random_state=78)
        grid = VisualizerGrid(visualizers)

        grid.fit(X, y)
        # show is required here (do not replace with finalize)!
        assert grid.show() is not None

        self.assert_images_similar(grid, tol=1.0)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_draw_with_rows(self):
        """
        Draw 2 visualizers in their own row
        """
        visualizers = [
            RandomVisualizer(random_state=63),
            RandomVisualizer(random_state=36),
        ]

        X, y = make_classification(random_state=87)
        grid = VisualizerGrid(visualizers, nrows=2)

        grid.fit(X, y)
        # show is required here (do not replace with finalize)!
        assert grid.show() is not None

        self.assert_images_similar(grid, tol=1.0)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_draw_with_cols(self):
        """
        Draw 2 visualizers in their own column
        """
        visualizers = [
            RandomVisualizer(random_state=633),
            RandomVisualizer(random_state=336),
        ]

        X, y = make_classification(random_state=187)
        grid = VisualizerGrid(visualizers, ncols=2)

        grid.fit(X, y)
        # show is required here (do not replace with finalize)!
        assert grid.show() is not None

        self.assert_images_similar(grid, tol=1.0)

    def test_cant_define_both_rows_cols(self):
        """
        Assert that both nrows and ncols cannot be specified
        """
        with pytest.raises(YellowbrickValueError):
            VisualizerGrid([], ncols=2, nrows=2)



# tests.test_regressor.test_alphas
# Tests for the alpha selection visualizations.
#
# Author:   Benjamin Bengfort
# Created:  Tue Mar 07 12:13:04 2017 -0500
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_alphas.py [7d3f5e6] benjamin@bengfort.com $

"""
Tests for the alpha selection visualizations.
"""

##########################################################################
## Imports
##########################################################################

import sys
import pytest
import numpy as np

from tests.base import VisualTestCase
from numpy.testing import assert_array_equal

from yellowbrick.datasets import load_energy, load_concrete
from yellowbrick.exceptions import YellowbrickTypeError
from yellowbrick.exceptions import YellowbrickValueError
from yellowbrick.regressor.alphas import AlphaSelection, alphas
from yellowbrick.regressor.alphas import ManualAlphaSelection, manual_alphas


from sklearn.svm import SVR, SVC
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.datasets import make_regression
from sklearn.linear_model import Ridge, RidgeCV
from sklearn.linear_model import Lasso, LassoCV
from sklearn.linear_model import LassoLars, LassoLarsCV
from sklearn.linear_model import ElasticNet, ElasticNetCV


##########################################################################
## Alpha Selection Tests
##########################################################################


class TestAlphaSelection(VisualTestCase):
    """
    Test the AlphaSelection visualizer
    """

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_similar_image(self):
        """
        Integration test with image similarity comparison
        """

        visualizer = AlphaSelection(LassoCV(random_state=0))

        X, y = make_regression(random_state=0)
        visualizer.fit(X, y)
        visualizer.finalize()

        self.assert_images_similar(visualizer)

    @pytest.mark.parametrize("model", [SVR, Ridge, Lasso, LassoLars, ElasticNet])
    def test_regressor_nocv(self, model):
        """
        Ensure only "CV" regressors are allowed
        """
        with pytest.raises(YellowbrickTypeError):
            AlphaSelection(model())

    @pytest.mark.parametrize("model", [RidgeCV, LassoCV, LassoLarsCV, ElasticNetCV])
    def test_regressor_cv(self, model):
        """
        Ensure "CV" regressors are allowed
        """
        try:
            AlphaSelection(model())
        except YellowbrickTypeError:
            pytest.fail("could not instantiate RegressorCV on alpha selection")

    @pytest.mark.parametrize("model", [SVC, KMeans, PCA])
    def test_only_regressors(self, model):
        """
        Assert AlphaSelection only works with regressors
        """
        with pytest.raises(YellowbrickTypeError):
            AlphaSelection(model())

    def test_store_cv_values(self):
        """
        Assert that store_cv_results/store_cv_values is true on RidgeCV
        """
        # sklearn 1.7+ uses store_cv_results, older versions use store_cv_values
        model = AlphaSelection(RidgeCV())
        assert getattr(model.estimator, "store_cv_results", None) or getattr(
            model.estimator, "store_cv_values", None
        )

        # Test with explicit parameter (try both old and new names)
        try:
            model = AlphaSelection(RidgeCV(store_cv_results=True))
            assert model.estimator.store_cv_results
        except TypeError:
            # Fallback to old parameter name for sklearn < 1.7
            model = AlphaSelection(RidgeCV(store_cv_values=True))
            assert model.estimator.store_cv_values

        # Test that we override False to True
        try:
            model = AlphaSelection(RidgeCV(store_cv_results=False))
            assert model.estimator.store_cv_results
        except TypeError:
            model = AlphaSelection(RidgeCV(store_cv_values=False))
            assert model.estimator.store_cv_values

    @pytest.mark.parametrize("model", [RidgeCV, LassoCV, ElasticNetCV])
    def test_get_alphas_param(self, model):
        """
        Assert that we can get the alphas from original CV models
        """
        alphas = np.logspace(-10, -2, 100)

        try:
            model = AlphaSelection(model(alphas=alphas))
            malphas = model._find_alphas_param()
            assert_array_equal(alphas, malphas)
        except YellowbrickValueError:
            pytest.fail("could not find alphas on {}".format(model.name))

    def test_get_alphas_param_lassolars(self):
        """
        Assert that we can get alphas from lasso lars.
        """
        X, y = make_regression()
        model = AlphaSelection(LassoLarsCV())
        model.fit(X, y)
        try:
            malphas = model._find_alphas_param()
            assert len(malphas) > 0
        except YellowbrickValueError:
            pytest.fail("could not find alphas on {}".format(model.name))

    @pytest.mark.parametrize("model", [RidgeCV, LassoCV, LassoLarsCV, ElasticNetCV])
    def test_get_errors_param(self, model):
        """
        Test known models we can get the cv errors for alpha selection
        """
        try:
            model = AlphaSelection(model())

            X, y = make_regression()
            model.fit(X, y)

            errors = model._find_errors_param()
            assert len(errors) > 0
        except YellowbrickValueError:
            pytest.fail("could not find errors on {}".format(model.name))

    def test_score(self):
        """
        Assert the score method returns an R2 value
        """
        visualizer = AlphaSelection(RidgeCV())

        X, y = make_regression(random_state=352)
        visualizer.fit(X, y)
        assert visualizer.score(X, y) == pytest.approx(0.9999780266590336)

    def test_quick_method(self):
        """
        Test the quick method producing a valid visualization
        """
        X, y = load_energy(return_dataset=True).to_numpy()

        visualizer = alphas(LassoCV(random_state=0), X, y, is_fitted=False, show=False)
        assert isinstance(visualizer, AlphaSelection)
        self.assert_images_similar(visualizer, tol=0.1)

    def test_within_pipeline(self):
        """
        Test that visualizer can be accessed within a sklearn pipeline
        """
        X, y = load_concrete()

        # Create a list of alphas to cross-validate against
        alphas = np.logspace(-10, 1, 400)

        model = Pipeline(
            [
                ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
                ("alpha", AlphaSelection(LassoCV(random_state=42, alphas=alphas))),
            ]
        )

        model.fit(X, y)
        model["alpha"].finalize()
        self.assert_images_similar(model["alpha"], tol=2.0)


class TestManualAlphaSelection(VisualTestCase):
    """
    Test the ManualAlphaSelection visualizer
    """

    def test_similar_image_manual(self):
        """
        Integration test with image similarity comparison
        """

        visualizer = ManualAlphaSelection(Lasso(random_state=0), cv=5)

        X, y = make_regression(random_state=0)
        visualizer.fit(X, y)
        visualizer.finalize()

        # Image comparison fails on Appveyor with RMS 0.024
        self.assert_images_similar(visualizer, tol=0.1)

    @pytest.mark.parametrize("model", [RidgeCV, LassoCV, LassoLarsCV, ElasticNetCV])
    def test_manual_with_cv(self, model):
        """
        Ensure only non-CV regressors are allowed
        """
        with pytest.raises(YellowbrickTypeError):
            ManualAlphaSelection(model())

    @pytest.mark.parametrize("model", [SVR, Ridge, Lasso, LassoLars, ElasticNet])
    def test_manual_no_cv(self, model):
        """
        Ensure non-CV regressors are allowed
        """
        try:
            ManualAlphaSelection(model())
        except YellowbrickTypeError:
            pytest.fail("could not instantiate Regressor on alpha selection")

    def test_quick_method_manual(self):
        """
        Test the manual alphas quick method producing a valid visualization
        """
        X, y = load_energy(return_dataset=True).to_numpy()

        visualizer = manual_alphas(
            ElasticNet(random_state=0), X, y, cv=3, is_fitted=False, show=False
        )
        assert isinstance(visualizer, ManualAlphaSelection)
        # Python 3.6 Travis images not similar (RMS 0.024)
        self.assert_images_similar(visualizer, tol=0.5)

    def test_manual_within_pipeline(self):
        """
        Test that visualizer can be accessed within a sklearn pipeline
        """
        X, y = load_concrete()

        # Create a list of alphas to cross-validate against
        alpha_values = np.logspace(1, 4, 50)

        model = Pipeline(
            [
                ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
                (
                    "alpha",
                    ManualAlphaSelection(
                        Ridge(random_state=42),
                        alphas=alpha_values,
                        cv=12,
                        scoring="neg_mean_squared_error",
                    ),
                ),
            ]
        )

        model.fit(X, y)
        model["alpha"].finalize()
        self.assert_images_similar(model["alpha"], tol=2.0)



# tests.test_model_selection.test_importances
# Test the feature importance visualizers
#
# Author:  Benjamin Bengfort
# Author:  Rebecca Bilbro
# Created: Fri Mar 02 15:23:22 2018 -0500
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_importances.py [] benjamin@bengfort.com $

"""
Test the feature importance visualizers
"""

##########################################################################
## Imports
##########################################################################

import pytest
import numpy as np
import numpy.testing as npt
import matplotlib.pyplot as plt

from yellowbrick.exceptions import NotFitted
from yellowbrick.model_selection.importances import *
from yellowbrick.datasets import load_occupancy, load_concrete

from sklearn.datasets import load_iris
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.linear_model import LogisticRegression, Lasso
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline

from unittest import mock
from tests.base import VisualTestCase

try:
    import pandas as pd
except ImportError:
    pd = None


##########################################################################
## Feature Importances Tests
##########################################################################


class TestFeatureImportancesVisualizer(VisualTestCase):
    """
    Test FeatureImportances visualizer
    """

    def test_integration_feature_importances(self):
        """
        Integration test of visualizer with feature importances param
        """

        # Load the test dataset
        X, y = load_occupancy(return_dataset=True).to_numpy()

        fig = plt.figure()
        ax = fig.add_subplot()

        clf = GradientBoostingClassifier(random_state=42)
        viz = FeatureImportances(clf, ax=ax)
        viz.fit(X, y)
        viz.finalize()

        # Appveyor and Linux conda non-text-based differences
        self.assert_images_similar(viz, tol=13.0)

    def test_integration_coef(self):
        """
        Integration test of visualizer with coef param
        """

        # Load the test dataset
        dataset = load_concrete(return_dataset=True)
        X, y = dataset.to_numpy()
        features = dataset.meta["features"]

        fig = plt.figure()
        ax = fig.add_subplot()

        reg = Lasso(random_state=42)
        features = list(map(lambda s: s.title(), features))
        viz = FeatureImportances(reg, ax=ax, labels=features, relative=False)
        viz.fit(X, y)
        viz.finalize()

        # Appveyor and Linux conda non-text-based differences
        self.assert_images_similar(viz, tol=16.2)

    def test_integration_quick_method(self):
        """
        Integration test of quick method
        """

        # Load the test dataset
        X, y = load_occupancy(return_dataset=True).to_numpy()

        fig = plt.figure()
        ax = fig.add_subplot()

        clf = RandomForestClassifier(random_state=42)
        g = feature_importances(clf, X, y, ax=ax, show=False)

        # Appveyor and Linux conda non-text-based differences
        self.assert_images_similar(g, tol=15.0)

    def test_fit_no_importances_model(self):
        """
        Fitting a model without feature importances raises an exception
        """
        X = np.random.rand(100, 42)
        y = np.random.rand(100)

        visualizer = FeatureImportances(MockEstimator())
        expected_error = "could not find feature importances param on MockEstimator"

        with pytest.raises(YellowbrickTypeError, match=expected_error):
            visualizer.fit(X, y)

    def test_fit_sorted_params(self):
        """
        On fit, sorted features_ and feature_importances_ params are created
        """
        coefs = np.array([0.4, 0.2, 0.08, 0.07, 0.16, 0.23, 0.38, 0.1, 0.05])
        names = np.array(["a", "b", "c", "d", "e", "f", "g", "h", "i"])

        model = MockEstimator()
        model.make_importance_param(value=coefs)

        visualizer = FeatureImportances(model, labels=names)
        visualizer.fit(np.random.rand(100, len(names)), np.random.rand(100))

        assert hasattr(visualizer, "features_")
        assert hasattr(visualizer, "feature_importances_")

        # get the expected sort index
        sort_idx = np.argsort(coefs)

        # assert sorted
        npt.assert_array_equal(names[sort_idx], visualizer.features_)
        npt.assert_array_equal(coefs[sort_idx], visualizer.feature_importances_)

    def test_fit_relative(self):
        """
        Test fit computes relative importances
        """
        coefs = np.array([0.4, 0.2, 0.08, 0.07, 0.16, 0.23, 0.38, 0.1, 0.05])

        model = MockEstimator()
        model.make_importance_param(value=coefs)

        visualizer = FeatureImportances(model, relative=True)
        visualizer.fit(np.random.rand(100, len(coefs)), np.random.rand(100))

        expected = 100.0 * coefs / coefs.max()
        expected.sort()
        npt.assert_array_equal(visualizer.feature_importances_, expected)

    def test_fit_not_relative(self):
        """
        Test fit stores unmodified importances
        """
        coefs = np.array([0.4, 0.2, 0.08, 0.07, 0.16, 0.23, 0.38, 0.1, 0.05])

        model = MockEstimator()
        model.make_importance_param(value=coefs)

        visualizer = FeatureImportances(model, relative=False)
        visualizer.fit(np.random.rand(100, len(coefs)), np.random.rand(100))

        coefs.sort()
        npt.assert_array_equal(visualizer.feature_importances_, coefs)

    def test_fit_absolute(self):
        """
        Test fit with absolute values
        """
        coefs = np.array([0.4, 0.2, -0.08, 0.07, 0.16, 0.23, -0.38, 0.1, -0.05])

        model = MockEstimator()
        model.make_importance_param(value=coefs)

        # Test absolute value
        visualizer = FeatureImportances(model, absolute=True, relative=False)
        visualizer.fit(np.random.rand(100, len(coefs)), np.random.rand(100))

        expected = np.array([0.05, 0.07, 0.08, 0.1, 0.16, 0.2, 0.23, 0.38, 0.4])
        npt.assert_array_equal(visualizer.feature_importances_, expected)

        # Test no absolute value
        visualizer = FeatureImportances(model, absolute=False, relative=False)
        visualizer.fit(np.random.rand(100, len(coefs)), np.random.rand(100))

        expected = np.array([-0.38, -0.08, -0.05, 0.07, 0.1, 0.16, 0.2, 0.23, 0.4])
        npt.assert_array_equal(visualizer.feature_importances_, expected)

    def test_multi_coefs(self):
        """
        Test fit with multidimensional coefficients and stack warning
        """
        coefs = np.array(
            [
                [0.4, 0.2, -0.08, 0.07, 0.16, 0.23, -0.38, 0.1, -0.05],
                [0.41, 0.12, -0.1, 0.1, 0.14, 0.21, 0.01, 0.31, -0.15],
                [0.31, 0.2, -0.01, 0.1, 0.22, 0.23, 0.01, 0.12, -0.15],
            ]
        )

        model = MockEstimator()
        model.make_importance_param(value=coefs)

        visualizer = FeatureImportances(model, stack=False)

        with pytest.warns(YellowbrickWarning):
            visualizer.fit(
                np.random.rand(100, len(np.mean(coefs, axis=0))), np.random.rand(100)
            )

        npt.assert_equal(visualizer.feature_importances_.ndim, 1)

    def test_multi_coefs_stacked(self):
        """
        Test stack plot with multidimensional coefficients
        """
        X, y = load_iris(return_X_y=True)

        viz = FeatureImportances(
            LogisticRegression(solver="liblinear", random_state=222), stack=True
        )
        viz.fit(X, y)
        viz.finalize()

        npt.assert_equal(viz.feature_importances_.shape, (3, 4))
        # Appveyor and Linux conda non-text-based differences
        self.assert_images_similar(viz, tol=17.5)

    def test_stack_param_incorrectly_used_throws_error(self):
        """
        Test incorrectly using stack param on a dataset with two classes which
        does not return a coef_ array in the shape of (n_classes, n_features)
        """
        X, y = load_occupancy()

        viz = FeatureImportances(
            LogisticRegression(solver="liblinear", random_state=222), stack=True
        )

        expected_error = "The model used does not return coef_ array"

        with pytest.raises(YellowbrickValueError, match=expected_error):
            viz.fit(X, y)

    @pytest.mark.skipif(pd is None, reason="pandas is required for this test")
    def test_fit_dataframe(self):
        """
        Ensure feature names are extracted from DataFrame columns
        """
        labels = ["a", "b", "c", "d", "e", "f"]
        df = pd.DataFrame(np.random.rand(100, 6), columns=labels)
        s = pd.Series(np.random.rand(100), name="target")

        assert df.shape == (100, 6)

        model = MockEstimator()
        model.make_importance_param(value=np.linspace(0, 1, 6))

        visualizer = FeatureImportances(model)
        visualizer.fit(df, s)

        assert hasattr(visualizer, "features_")
        npt.assert_array_equal(visualizer.features_, np.array(df.columns))

    def test_fit_makes_labels(self):
        """
        Assert that the fit process makes label indices
        """
        model = MockEstimator()
        model.make_importance_param(value=np.linspace(0, 1, 10))

        visualizer = FeatureImportances(model)
        visualizer.fit(np.random.rand(100, 10), np.random.rand(100))

        # Don't have to worry about label space since importances are linspace
        assert hasattr(visualizer, "features_")
        npt.assert_array_equal(np.arange(10), visualizer.features_)

    def test_fit_calls_draw(self):
        """
        Assert that fit calls draw
        """
        model = MockEstimator()
        model.make_importance_param("coef_")

        visualizer = FeatureImportances(model)

        with mock.patch.object(visualizer, "draw") as mdraw:
            visualizer.fit(np.random.rand(100, 42), np.random.rand(100))
            mdraw.assert_called_once()

    def test_draw_raises_unfitted(self):
        """
        Assert draw raises exception when not fitted
        """
        visualizer = FeatureImportances(Lasso())
        with pytest.raises(NotFitted):
            visualizer.draw()

    def test_find_importances_param(self):
        """
        Test the expected parameters can be found
        """
        params = ("feature_importances_", "coef_")

        for param in params:
            model = MockEstimator()
            model.make_importance_param(param, "foo")
            visualizer = FeatureImportances(model)

            assert hasattr(model, param), "expected '{}' missing".format(param)
            for oparam in params:
                if oparam == param:
                    continue
                assert not hasattr(model, oparam), "unexpected '{}'".format(oparam)

            importances = visualizer._find_importances_param()
            assert importances == "foo"

    def test_find_importances_param_priority(self):
        """
        With both feature_importances_ and coef_, one has priority
        """
        model = MockEstimator()
        model.make_importance_param("feature_importances_", "foo")
        model.make_importance_param("coef_", "bar")
        visualizer = FeatureImportances(model)

        assert hasattr(model, "feature_importances_")
        assert hasattr(model, "coef_")

        importances = visualizer._find_importances_param()
        assert importances == "foo"

    def test_find_importances_param_not_found(self):
        """
        Raises an exception when importances param not found
        """
        model = MockEstimator()
        visualizer = FeatureImportances(model)

        assert not hasattr(model, "feature_importances_")
        assert not hasattr(model, "coef_")

        with pytest.raises(YellowbrickTypeError):
            visualizer._find_importances_param()

    def test_find_classes_param_not_found(self):
        """
        Raises an exception when classes param not found
        """
        model = MockClassifier()
        visualizer = FeatureImportances(model)

        assert not hasattr(model, "classes_")

        e = "could not find classes_ param on {}".format(
            visualizer.estimator.__class__.__name__
        )
        with pytest.raises(YellowbrickTypeError, match=e):
            visualizer._find_classes_param()

    def test_xlabel(self):
        """
        Check the various xlabels are sensical
        """
        model = MockEstimator()
        model.make_importance_param("feature_importances_")
        visualizer = FeatureImportances(model, xlabel="foo", relative=True)

        # Assert the visualizer uses the user supplied xlabel
        assert visualizer._get_xlabel() == "foo", "could not set user xlabel"

        # Check the visualizer default relative xlabel
        visualizer.set_params(xlabel=None)
        assert "relative" in visualizer._get_xlabel()

        # Check value xlabel with default
        visualizer.set_params(relative=False)
        assert "relative" not in visualizer._get_xlabel()

        # Check coeficients
        model = MockEstimator()
        model.make_importance_param("coef_")
        visualizer = FeatureImportances(model, xlabel="baz", relative=True)

        # Assert the visualizer uses the user supplied xlabel
        assert visualizer._get_xlabel() == "baz", "could not set user xlabel"

        # Check the visualizer default relative xlabel
        visualizer.set_params(xlabel=None)
        assert "coefficient" in visualizer._get_xlabel()
        assert "relative" in visualizer._get_xlabel()

        # Check value xlabel with default
        visualizer.set_params(relative=False)
        assert "coefficient" in visualizer._get_xlabel()
        assert "relative" not in visualizer._get_xlabel()

    def test_is_fitted(self):
        """
        Test identification if is fitted
        """
        visualizer = FeatureImportances(Lasso())
        assert not visualizer._is_fitted()

        visualizer.features_ = "foo"
        assert not visualizer._is_fitted()

        visualizer.feature_importances_ = "bar"
        assert visualizer._is_fitted()

        del visualizer.features_
        assert not visualizer._is_fitted()

    def test_with_fitted(self):
        """
        Test that visualizer properly handles an already-fitted model
        """
        X, y = load_concrete(return_dataset=True).to_numpy()

        model = Lasso().fit(X, y)

        with mock.patch.object(model, "fit") as mockfit:
            oz = FeatureImportances(model)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with mock.patch.object(model, "fit") as mockfit:
            oz = FeatureImportances(model, is_fitted=True)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with mock.patch.object(model, "fit") as mockfit:
            oz = FeatureImportances(model, is_fitted=False)
            oz.fit(X, y)
            mockfit.assert_called_once_with(X, y)

    def test_topn_stacked(self):
        """
        Test stack plot with only the three most important features by sum of
        each feature's importance across all classes
        """
        X, y = load_iris(return_X_y=True)

        viz = FeatureImportances(
            LogisticRegression(solver="liblinear", random_state=222),
            stack=True, topn=3
        )
        viz.fit(X, y)
        viz.finalize()

        npt.assert_equal(viz.feature_importances_.shape, (3, 3))
        # Appveyor and Linux conda non-text-based differences
        self.assert_images_similar(viz, tol=17.5)

    def test_topn_negative_stacked(self):
        """
        Test stack plot with only the three least important features by sum of
        each feature's importance across all classes
        """
        X, y = load_iris(return_X_y=True)

        viz = FeatureImportances(
            LogisticRegression(solver="liblinear", random_state=222),
            stack=True, topn=-3
        )
        viz.fit(X, y)
        viz.finalize()

        npt.assert_equal(viz.feature_importances_.shape, (3, 3))
        # Appveyor and Linux conda non-text-based differences
        self.assert_images_similar(viz, tol=17.5)

    def test_topn(self):
        """
        Test plot with only top three important features by absolute value
        """
        X, y = load_iris(return_X_y=True)

        viz = FeatureImportances(
            GradientBoostingClassifier(random_state=42), topn=3
        )
        viz.fit(X, y)
        viz.finalize()

        # Appveyor and Linux conda non-text-based differences
        self.assert_images_similar(viz, tol=17.5)

    def test_topn_negative(self):
        """
        Test plot with only the three least important features by absolute value
        """
        X, y = load_iris(return_X_y=True)

        viz = FeatureImportances(
            GradientBoostingClassifier(random_state=42), topn=-3
        )
        viz.fit(X, y)
        viz.finalize()

        # Appveyor and Linux conda non-text-based differences
        self.assert_images_similar(viz, tol=17.5)

    def test_within_pipeline(self):
        """
        Test that visualizer can be accessed within a sklearn pipeline
        """
        dataset = load_concrete(return_dataset=True)
        X, y = dataset.to_data()
        features = dataset.meta["features"]
        features = list(map(lambda s: s.title(), features))

        model = Pipeline([
            ('imputer', SimpleImputer(missing_values=np.nan, strategy='mean')),
            ('fi', FeatureImportances(Lasso(random_state=42), labels=features, relative=False))
        ])

        model.fit(X, y)
        model['fi'].finalize()
        self.assert_images_similar(model['fi'], tol=17.5)

    def test_within_pipeline_quickmethod(self):
        """
        Test that visualizer quickmethod can be accessed within a
        sklearn pipeline
        """
        dataset = load_concrete(return_dataset=True)
        X, y = dataset.to_data()
        features = dataset.meta["features"]
        features = list(map(lambda s: s.title(), features))

        model = Pipeline([
            ('imputer', SimpleImputer(missing_values=np.nan, strategy='mean')),
            ('fi', feature_importances(Lasso(random_state=42), X, y, labels=features, relative=False, show=False))
        ])
        self.assert_images_similar(model['fi'], tol=17.5)



##########################################################################
## Mock Estimator
##########################################################################


class MockEstimator(BaseEstimator):
    """
    Creates params when fit is called on demand.
    """

    def make_importance_param(self, name="feature_importances_", value=None):
        if value is None:
            value = np.random.rand(42)
        setattr(self, name, value)

    def fit(self, X, y=None, **kwargs):
        return self


class MockClassifier(BaseEstimator, ClassifierMixin):
    """
    Creates empty classifier.
    """

    pass



# -*- coding: utf-8 -*-
# tests.test_utils.test_helpers
# Tests for the stand alone helper functions in Yellowbrick utils.
#
# Author:   Benjamin Bengfort
# Author:   Rebecca Bilbro
# Created:  Fri May 19 10:43:43 2017 -0700
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_helpers.py [79cd8cf] benjamin@bengfort.com $

"""
Tests for the stand alone helper functions in Yellowbrick utils.
"""

##########################################################################
## Imports
##########################################################################

import pytest
import numpy as np
import numpy.testing as npt

from sklearn.svm import SVR, SVC
from sklearn.pipeline import Pipeline
from sklearn.decomposition import PCA
from sklearn.naive_bayes import GaussianNB
from sklearn.tree import DecisionTreeClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.cluster import Birch, AgglomerativeClustering
from sklearn.decomposition import LatentDirichletAllocation as LDA
from sklearn.cluster import KMeans, AffinityPropagation, MiniBatchKMeans
from sklearn.datasets import make_classification, make_regression, make_blobs
from sklearn.linear_model import Ridge, RidgeCV, Lasso, LassoCV, LinearRegression

from yellowbrick.utils.helpers import *

##########################################################################
## Helper Function Tests
##########################################################################


class TestHelpers(object):
    """
    Helper functions and utilities
    """

    @pytest.mark.parametrize(
        "model, name",
        [
            (LassoCV, "LassoCV"),
            (KNeighborsClassifier, "KNeighborsClassifier"),
            (KMeans, "KMeans"),
            (RandomForestClassifier, "RandomForestClassifier"),
        ],
        ids=["LassoCV", "KNeighborsClassifier", "KMeans", "RandomForestClassifier"],
    )
    def test_real_model(self, model, name):
        """
        Test getting model name for sklearn estimators
        """
        assert get_model_name(model()) == name

    def test_pipeline(self):
        """
        Test getting model name for sklearn pipelines
        """
        pipeline = Pipeline([("reduce_dim", PCA()), ("linreg", LinearRegression())])
        assert get_model_name(pipeline) == "LinearRegression"

    def test_int_input(self):
        """
        Assert a type error is raised when an int is passed to model name
        """
        with pytest.raises(TypeError):
            get_model_name(1)

    def test_str_input(self):
        """
        Assert a type error is raised when a str is passed to model name
        """
        with pytest.raises(TypeError):
            get_model_name("helloworld")

    @pytest.mark.parametrize(
        "regressor",
        [SVR, Ridge, Lasso, RidgeCV, LassoCV],
        ids=["SVR", "Ridge", "Lasso", "RidgeCV", "LassoCV"],
    )
    def test_is_fitted_regressors(self, regressor):
        """
        Assert sklearn regressors return False for fitted before fit and True after
        """
        assert is_fitted(regressor()) is False

        X, y = make_regression(n_samples=20, n_features=3)
        fitted = regressor().fit(X, y)
        assert is_fitted(fitted) is True

    @pytest.mark.parametrize(
        "classifier",
        [
            SVC,
            GaussianNB,
            KNeighborsClassifier,
            RandomForestClassifier,
            DecisionTreeClassifier,
        ],
        ids=[
            "SVC",
            "GaussianNB",
            "KNeighborsClassifier",
            "RandomForestClassifier",
            "DecisionTreeClassifier",
        ],
    )
    def test_is_fitted_classifiers(self, classifier):
        """
        Assert sklearn classifiers return False for fitted before fit and True after
        """
        assert is_fitted(classifier()) is False

        X, y = make_classification(n_samples=20, n_features=3, n_redundant=0)
        fitted = classifier().fit(X, y)
        assert is_fitted(fitted) is True

    @pytest.mark.parametrize(
        "clusterer",
        [
            PCA,  # AttributeError: has no attribute 'predict'
            LDA,  # AttributeError: has no attribute 'predict'
            Birch,
            KMeans,
            MiniBatchKMeans,
            AffinityPropagation,
            AgglomerativeClustering,  # AttributeError: has no attribute 'predict'
        ],
        ids=[
            "PCA",
            "LDA",
            "Birch",
            "KMeans",
            "MiniBatchKMeans",
            "AffinityPropagation",
            "AgglomerativeClustering",
        ],
    )
    def test_is_fitted_clusterers(self, clusterer):
        """
        Assert sklearn clusterers return False for fitted before fit and True after
        """
        assert is_fitted(clusterer()) is False

        X, _ = make_blobs(n_samples=20, center_box=(10.0, 20.0))
        fitted = clusterer().fit(X)
        assert is_fitted(fitted) is True

    def test_check_fitted(self):
        """
        Verify the user can specify whether or not to check if the model is fitted
        """
        model = SVC()
        assert check_fitted(model, is_fitted_by="auto") is False
        assert check_fitted(model, is_fitted_by=True) is True
        assert check_fitted(model, is_fitted_by=False) is False

        X, y = make_classification(n_samples=20, n_features=3, n_redundant=0)
        model.fit(X, y)
        assert check_fitted(model, is_fitted_by="auto") is True
        assert check_fitted(model, is_fitted_by=True) is True
        assert check_fitted(model, is_fitted_by=False) is False

    @pytest.mark.parametrize(
        "estimator",
        [
            SVC,
            SVR,
            Ridge,
            KMeans,
            RidgeCV,
            GaussianNB,
            MiniBatchKMeans,
            LinearRegression,
        ],
        ids=[
            "SVC",
            "SVR",
            "Ridge",
            "KMeans",
            "RidgeCV",
            "GaussianNB",
            "MiniBatchKMeans",
            "LinearRegression",
        ],
    )
    def test_get_param_names(self, estimator):
        """
        Assert we successfully extract the parameters from sklearn estimators
        """
        assert "sample_weight" in get_param_names(estimator.fit)

    def test_get_param_names_type(self):
        """
        Assert a type error is raised when passing a non-method
        """
        with pytest.raises(TypeError):
            get_param_names("test")


##########################################################################
## Numeric Function Tests
##########################################################################


class TestNumericFunctions(object):
    """
    Numeric helper functions
    """

    def test_div_1d_by_scalar(self):
        """
        Test divide 1D vector by scalar
        """
        result = div_safe([-1, 0, 1], 0)
        assert result.all() == 0

    def test_div_1d_by_1d(self):
        """
        Test divide 1D vector by another 1D vector with same length
        """
        result = div_safe([-1, 0, 1], [0, 0, 0])
        assert result.all() == 0

    def test_div_2d_by_1d(self):
        """
        Test divide 2D vector by 1D vector with similar shape component
        """
        numerator = np.array([[-1, 0, 1, 2], [1, -1, 0, 3]])
        denominator = [0, 0, 0, 0]
        npt.assert_array_equal(
            div_safe(numerator, denominator), np.array([[0, 0, 0, 0], [0, 0, 0, 0]])
        )

    def test_invalid_dimensions(self):
        """
        Assert an error is raised on division with invalid dimensions
        """
        numerator = np.array([[-1, 0, 1, 2], [1, -1, 0, 3]])
        denominator = [0, 0]
        with pytest.raises(ValueError):
            div_safe(numerator, denominator)

    def test_div_scalar_by_scalar(self):
        """
        Assert a value error is raised when trying to divide two scalars
        """
        with pytest.raises(ValueError):
            div_safe(5, 0)

    def test_prop_to_size_list(self):
        """
        Test prop to size correctly returns scaled values for a list
        """
        # Hieghts (in cm) of U.S. Presidents in order of term until Lincoln
        heights = [
            188,
            170,
            189,
            163,
            183,
            171,
            185,
            168,
            173,
            183,
            173,
            173,
            175,
            178,
            183,
            193,
        ]
        sizes = prop_to_size(heights, mi=1, ma=10, log=False, power=0.33)

        npt.assert_array_almost_equal(
            sizes,
            np.array(
                [
                    9.47447296,
                    6.56768746,
                    9.58486955,
                    1.0,
                    8.87285756,
                    6.81851544,
                    9.12441277,
                    5.98256068,
                    7.26314542,
                    8.87285756,
                    7.26314542,
                    7.26314542,
                    7.65154152,
                    8.15982835,
                    8.87285756,
                    10.0,
                ]
            ),
        )

    def test_prop_to_size_log(self):
        """
        Test prop to size returns natural log scaled values
        """
        # Heights (in cm) of U.S. Presidents in order of term until Lincoln
        heights = [
            188,
            170,
            189,
            163,
            183,
            171,
            185,
            168,
            173,
            183,
            173,
            173,
            175,
            178,
            183,
            193,
        ]
        sizes = prop_to_size(heights, mi=1, ma=10, log=True, power=0.5)

        npt.assert_array_almost_equal(
            sizes,
            np.array(
                [
                    9.271337,
                    5.49004,
                    9.423692,
                    1.0,
                    8.449214,
                    5.792968,
                    8.791172,
                    4.806088,
                    6.343007,
                    8.449214,
                    6.343007,
                    6.343007,
                    6.835994,
                    7.496806,
                    8.449214,
                    10.0,
                ]
            ),
        )

    def test_prop_to_size_default(self):
        """
        Test the default values of prop to size are correct
        """
        vals = np.random.normal(50, 23, 500)
        sizes = prop_to_size(vals)

        assert sizes.ndim == vals.ndim
        assert sizes.shape == vals.shape
        assert sizes.max() <= 5.0
        assert sizes.min() >= 0.0

    def test_prop_to_size_zero_division(self):
        """
        Ensure that prop to size does not cause division by zero errors
        """
        vals = [8] * 8
        sizes = prop_to_size(vals)
        npt.assert_array_equal(sizes, [0] * 8)


##########################################################################
## Features/Array Tests
##########################################################################


class TestNarrayIntColumns(object):
    """
    Features and array helper tests
    """

    def test_has_ndarray_int_columns_true_int_features(self):
        """
        Ensure ndarray with int features has int columns
        """
        x = np.random.rand(3, 5)
        features = [0, 1]
        assert has_ndarray_int_columns(features, x)

    def test_has_ndarray_int_columns_true_int_strings(self):
        """
        Ensure ndarray with str(int) features has int columns
        """
        x = np.random.rand(3, 5)
        features = ["0", "1"]
        assert has_ndarray_int_columns(features, x)

    def test_has_ndarray_int_columns_false_not_numeric(self):
        """
        Ensure ndarray with str features does not have int columns
        """
        x = np.random.rand(3, 5)
        features = ["a", "1"]
        assert not has_ndarray_int_columns(features, x)

    def test_has_ndarray_int_columns_false_outside_column_range(self):
        """
        Ensure ndarray with str(int) outside range does not have int columns
        """
        x = np.random.rand(3, 5)
        features = ["0", "10"]
        assert not has_ndarray_int_columns(features, x)

    @pytest.mark.parametrize(
        "a, increasing",
        [
            (np.array([0.8]), True),
            (np.array([9]), False),
            (np.array([0.2, 1.3, 1.4, 1.4, 1.4, 1.5, 8.3, 8.5]), True),
            (np.array([8, 7, 6, 5, 5, 5, 5, 4, 3, -1, -5]), False),
        ],
        ids=["increasing single", "decreasing single", "increasing", "decreasing"],
    )
    def test_is_monotonic(self, a, increasing):
        """
        Test if a vector is monotonic
        """
        assert is_monotonic(a, increasing)

    @pytest.mark.parametrize(
        "a, increasing",
        [
            (np.array([0.2, 1.3, 1.3, 0.2, 1.8]), True),
            (np.array([8, 7, 7, 8, 9, 6, 5]), False),
        ],
        ids=["increasing", "decreasing"],
    )
    def test_not_is_monotonic(self, a, increasing):
        """
        Test if a vector is not monotonic
        """
        assert not is_monotonic(a, increasing)

    def test_multi_dim_is_monotonic(self):
        """
        Assert monotonicity is not decidable on multi-dimensional array
        """
        with pytest.raises(ValueError):
            is_monotonic(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))


##########################################################################
## String Helpers Tests
##########################################################################


class TestStringHelpers(object):
    """
    String helper functions
    """

    def test_slugifiy(self):
        """
        Test the slugify helper utility
        """

        cases = (
            ("This is a test ---", "this-is-a-test"),
            ("This -- is a ## test ---", "this-is-a-test"),
        )

        for case, expected in cases:
            assert expected == slugify(case)



#!/usr/bin/env python3
# rfecv.py
# Generates RFECV visualizations for the documentation
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: rfecv.py [] $

import os
import matplotlib.pyplot as plt

from sklearn.svm import SVC
from sklearn.datasets import make_classification
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import StratifiedKFold

from yellowbrick.model_selection import RFECV, rfecv
from yellowbrick.datasets import load_credit


CWD = os.path.dirname(__file__)
IMAGES = os.path.join(CWD, "images")


def rfecv_sklearn_example(image="rfecv_sklearn_example.png"):
    X, y = make_classification(
        n_samples=1000,
        n_features=25,
        n_informative=3,
        n_redundant=2,
        n_repeated=0,
        n_classes=8,
        n_clusters_per_class=1,
        random_state=0,
    )

    _, ax = plt.subplots()

    oz = RFECV(SVC(kernel="linear", C=1), ax=ax)
    oz.fit(X, y)
    oz.show(outpath=os.path.join(IMAGES, image))


def rfecv_credit_example(image="rfecv_credit.png"):
    X, y = load_credit()

    _, ax = plt.subplots()
    cv = StratifiedKFold(5)
    oz = RFECV(RandomForestClassifier(), ax=ax, cv=cv, scoring="f1_weighted")
    oz.fit(X, y)
    oz.show(outpath=os.path.join(IMAGES, image))


def rfecv_quick_method(image="rfecv_quick_method.png"):
    X, y = load_credit()

    _, ax = plt.subplots()
    cv = StratifiedKFold(5)
    visualizer = rfecv(RandomForestClassifier(), X=X, y=y, ax=ax, cv=cv, scoring='f1_weighted')
    visualizer.show(outpath=os.path.join(IMAGES, image))


##########################################################################
## Main Method
##########################################################################

if __name__ == "__main__":
    rfecv_sklearn_example()
    rfecv_credit_example()
    rfecv_quick_method()



import pandas as pd
import csv
from sklearn import neighbors
from sklearn import datasets
import numpy as np
import yellowbrick as yb
from yellowbrick.neighbors import KnnDecisionBoundariesVisualizer
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap


def load_adm_sat_school_data(return_X_y=False):

    with open("./merged_adm_sat_data.csv") as csv_file:
        data_file = csv.reader(csv_file)
        temp = next(data_file)
        n_samples = int(temp[0])
        n_features = int(temp[1])
        target_names = np.array(temp[2:])

    df = pd.read_csv(
        "./merged_adm_sat_data.csv", sep=",", usecols=(0, 1, 2, 3), skiprows=0
    )
    data = np.empty((n_samples, n_features), dtype=int)
    target = np.ma.empty((n_samples,), dtype=int)

    for index, row in df.iterrows():
        data[index] = np.asarray(
            [df.iloc[index][0], df.iloc[index][1], df.iloc[index][2]], dtype=np.float
        )
        target[index] = np.asarray(df.iloc[index][3], dtype=np.int)

    feature_names = np.array(["ACT_AVG", "SAT_AVG", "GRAD_DEBT", "REGION"])

    if return_X_y:
        return data, target

    return datasets.base.Bunch(
        data=data,
        target=target,
        target_names=target_names,
        DESCR="School Data set",
        feature_names=feature_names,
    )


def show_plot(X, y, n_neighbors=10, h=0.2):
    # Create color maps
    cmap_light = ListedColormap(
        [
            "#FFAAAA",
            "#AAFFAA",
            "#AAAAFF",
            "#FFAAAA",
            "#AAFFAA",
            "#AAAAFF",
            "#FFAAAA",
            "#AAFFAA",
            "#AAAAFF",
            "#AAAAFF",
        ]
    )
    cmap_bold = ListedColormap(
        [
            "#FF0000",
            "#00FF00",
            "#0000FF",
            "#FF0000",
            "#FF0000",
            "#FF0000",
            "#FF0000",
            "#FF0000",
            "#FF0000",
            "#FF0000",
        ]
    )

    for weights in ["uniform", "distance"]:
        # we create an instance of Neighbours Classifier and fit the data.
        clf = neighbors.KNeighborsClassifier(n_neighbors, weights=weights)
        clf.fit(X, y)
        clf.n_neighbors = n_neighbors

        # Plot the decision boundary. For that, we will assign a color to each
        # point in the mesh [x_min, x_max]x[y_min, y_max].
        x_min, x_max = X[:, 0].min() - 1, X[:, 0].max() + 1
        y_min, y_max = X[:, 1].min() - 1, X[:, 1].max() + 1
        xx, yy = np.meshgrid(np.arange(x_min, x_max, h), np.arange(y_min, y_max, h))
        Z = clf.predict(np.c_[xx.ravel(), yy.ravel()])

        # Put the result into a color plot
        Z = Z.reshape(xx.shape)
        plt.figure()
        plt.pcolormesh(xx, yy, Z, cmap=cmap_light)

        # Plot also the training points
        plt.scatter(X[:, 0], X[:, 1], c=y, cmap=cmap_bold)
        plt.xlim(xx.min(), xx.max())
        plt.ylim(yy.min(), yy.max())
        plt.title(
            "3-Class classification (k = %i, weights = '%s')" % (n_neighbors, weights)
        )

    plt.show()


if __name__ == "__main__":
    school = load_adm_sat_school_data()
    X = school.data[:, :2]  # we only take the first two features.
    y = school.target
    # show_plot(X,y,3)
    model = neighbors.KNeighborsClassifier(10)
    model.fit(X, y)
    model.predict(X)
    # visualizer = KnnDecisionBoundariesVisualizer(model, classes=school.target_names, features=school.feature_names[:2])
    visualizer = KnnDecisionBoundariesVisualizer(model)
    visualizer.fit_draw_show(X, y)



# tests.test_classifier.test_class_prediction_error
# Testing for the ClassPredictionError visualizer
#
# Author:   Benjamin Bengfort
# Author:   Rebecca Bilbro
# Author:   Larry Gray
# Created:  Tue May 23 13:41:55 2017 -0700
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_rocauc.py [] benjamin@bengfort.com $

"""
Testing for the ClassPredictionError visualizer
"""

##########################################################################
## Imports
##########################################################################

import pytest
import matplotlib.pyplot as plt

from yellowbrick.exceptions import ModelError
from yellowbrick.datasets import load_occupancy
from yellowbrick.classifier.class_prediction_error import *

from sklearn.svm import LinearSVC, SVC
from sklearn.ensemble import RandomForestClassifier
from sklearn.datasets import make_multilabel_classification
from sklearn.model_selection import train_test_split as tts
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler

from unittest.mock import patch
from tests.base import VisualTestCase


try:
    import pandas as pd
except ImportError:
    pd = None


##########################################################################
##  Tests
##########################################################################


class TestClassPredictionError(VisualTestCase):
    """
    Test ClassPredictionError visualizer
    """

    @pytest.mark.filterwarnings("ignore:could not determine class_counts_")
    def test_numpy_integration(self):
        """
        Assert no errors during class prediction error integration with NumPy arrays
        """
        X, y = load_occupancy(return_dataset=True).to_numpy()

        classes = ["unoccupied", "occupied"]

        model = SVC(random_state=42)
        model.fit(X, y)
        visualizer = ClassPredictionError(model, classes=classes)
        visualizer.score(X, y)
        visualizer.finalize()

        # AppVeyor and Linux conda fail due to non-text-based differences
        # AppVeyor fails with RMS 13.161 - 13.289 (python - miniconda)
        self.assert_images_similar(visualizer, tol=12.5, windows_tol=13.3)

    @pytest.mark.filterwarnings("ignore:could not determine class_counts_")
    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_pandas_integration(self):
        """
        Assert no errors during class prediction error integration with Pandas
        """
        X, y = load_occupancy(return_dataset=True).to_pandas()
        classes = ["unoccupied", "occupied"]

        model = SVC(random_state=42)
        model.fit(X, y)
        visualizer = ClassPredictionError(model, classes=classes)
        visualizer.score(X, y)
        visualizer.finalize()

        # AppVeyor and Linux conda fail due to non-text-based differences
        # AppVeyor fails with RMS 13.161 - 13.289 (python - miniconda)
        self.assert_images_similar(visualizer, tol=12.5, windows_tol=13.3)

    def test_class_prediction_error_quickmethod(self):
        """
        Test the ClassPredictionError quickmethod
        """
        X, y = load_occupancy(return_dataset=True).to_numpy()

        fig = plt.figure()
        ax = fig.add_subplot()

        clf = SVC(random_state=42)
        viz = class_prediction_error(clf, X, y, ax=ax, show=False)

        # Not sure why the tolerance must be so high for this
        # Failing on travis with RMS 9.544
        # AppVeyor and Linux conda fail due to non-text-based differences: RMS 12.961
        # yellowbrick.exceptions.ImageComparisonFailure: images not close (RMS 15.538)
        self.assert_images_similar(viz, tol=16, windows_tol=16)

    def test_class_prediction_error_quickmethod_X_test_only(self):
        """
        Test the ClassPredictionError quickmethod
        """
        X, y = load_occupancy(return_dataset=True).to_numpy()
        X_train, X_test, y_train, y_test = tts(
            X, y, test_size=0.2, shuffle=True, random_state=42
        )

        fig = plt.figure()
        ax = fig.add_subplot()

        clf = LinearSVC(random_state=42)
        with pytest.raises(
            YellowbrickValueError,
            match="must specify both X_test and y_test or neither",
        ):
            class_prediction_error(
                clf, X_train=X_train, y_train=y_train, X_test=X_test, ax=ax, show=False
            )

    def test_class_prediction_error_quickmethod_X_test_and_y_test(self):
        """
        Test the ClassPredictionError quickmethod
        """
        X, y = load_occupancy(return_dataset=True).to_numpy()
        X_train, X_test, y_train, y_test = tts(
            X, y, test_size=0.2, shuffle=True, random_state=42
        )

        fig = plt.figure()
        ax = fig.add_subplot()

        clf = SVC(random_state=42)
        viz = class_prediction_error(
            clf,
            X_train=X_train,
            y_train=y_train,
            X_test=X_test,
            y_test=y_test,
            ax=ax,
            show=False,
        )

        # Not sure why the tolerance must be so high for this
        # Failing on travis with RMS 9.544
        # AppVeyor and Linux conda fail due to non-text-based differences: RMS 12.961
        self.assert_images_similar(viz, tol=13, windows_tol=13)

    @pytest.mark.filterwarnings("ignore:could not determine class_counts_")
    def test_classes_greater_than_indices(self):
        """
        A model error should be raised when there are more classes in fit than score
        """
        X, y = load_occupancy(return_dataset=True).to_numpy()
        classes = ["unoccupied", "occupied", "partytime"]

        model = LinearSVC(random_state=42)
        model.fit(X, y)
        with pytest.raises(ModelError):
            visualizer = ClassPredictionError(model, classes=classes)
            visualizer.score(X, y)

    def test_classes_less_than_indices(self):
        """
        Assert error when there is an attempt to filter classes
        """
        X, y = load_occupancy(return_dataset=True).to_numpy()
        classes = ["unoccupied"]

        model = LinearSVC(random_state=42)
        model.fit(X, y)
        with pytest.raises(NotImplementedError):
            visualizer = ClassPredictionError(model, classes=classes)
            visualizer.score(X, y)

    @pytest.mark.skip(reason="not implemented yet")
    def test_no_classes_provided(self):
        """
        Assert no errors when no classes are provided
        """
        pass

    def test_class_type(self):
        """
        Test class must be either binary or multiclass type
        """
        X, y = make_multilabel_classification()
        model = RandomForestClassifier()
        model.fit(X, y)
        with pytest.raises(YellowbrickValueError):
            visualizer = ClassPredictionError(model)
            visualizer.score(X, y)

    def test_score_returns_score(self):
        """
        Test that ClassPredictionError score() returns a score between 0 and 1
        """
        X, y = load_occupancy(return_dataset=True).to_numpy()

        # Create and fit the visualizer
        visualizer = ClassPredictionError(LinearSVC(random_state=42))
        visualizer.fit(X, y)

        # Score the visualizer
        s = visualizer.score(X, y)
        assert 0 <= s <= 1

    def test_with_fitted(self):
        """
        Test that visualizer properly handles an already-fitted model
        """
        X, y = load_occupancy(return_dataset=True).to_numpy()

        model = RandomForestClassifier().fit(X, y)
        classes = ["unoccupied", "occupied"]

        with patch.object(model, "fit") as mockfit:
            oz = ClassPredictionError(model, classes=classes)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with patch.object(model, "fit") as mockfit:
            oz = ClassPredictionError(model, classes=classes, is_fitted=True)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with patch.object(model, "fit") as mockfit:
            oz = ClassPredictionError(model, classes=classes, is_fitted=False)
            oz.fit(X, y)
            mockfit.assert_called_once_with(X, y)
    
    def test_within_pipeline(self):
        """
        Test that visualizer can be accessed within a sklearn pipeline
        """
        X, y = load_occupancy(return_dataset=True).to_pandas()
        classes = ["unoccupied", "occupied"]

        X_train, X_test, y_train, y_test = tts(
                    X, y, test_size=0.2, shuffle=True, random_state=42
                )

        model = Pipeline([
            ('minmax', MinMaxScaler()), 
            ('cpe', ClassPredictionError(SVC(random_state=42), classes=classes))
        ])

        model.fit(X_train, y_train)
        model.score(X_test, y_test)
        model['cpe'].finalize()
        self.assert_images_similar(model['cpe'], tol=12.5, windows_tol=13.3)

    def test_within_pipeline_quickmethod(self):
        """
        Test that visualizer quickmethod can be accessed within a
        sklearn pipeline
        """
        X, y = load_occupancy(return_dataset=True).to_pandas()

        X_train, X_test, y_train, y_test = tts(
                    X, y, test_size=0.2, shuffle=True, random_state=42
                )

        model = Pipeline([
            ('minmax', MinMaxScaler()), 
            ('cpe', class_prediction_error(SVC(random_state=42),
                                            X_train, y_train, X_test, y_test,
                                            classes=["vacant", "occupied"], show=False))
        ])
        self.assert_images_similar(model['cpe'], tol=12.5, windows_tol=13.3)

    def test_pipeline_as_model_input(self):
        """
        Test that visualizer can handle sklearn pipeline as model input
        """
        X, y = load_occupancy(return_dataset=True).to_pandas()
        classes = ["unoccupied", "occupied"]

        X_train, X_test, y_train, y_test = tts(
                    X, y, test_size=0.2, shuffle=True, random_state=42
                )

        model = Pipeline([
            ('minmax', MinMaxScaler()), 
            ('svc', SVC(random_state=42))
        ])

        oz = ClassPredictionError(model, classes=classes)
        oz.fit(X_train, y_train)
        oz.score(X_test, y_test)
        oz.finalize()
        self.assert_images_similar(oz, tol=12.5, windows_tol=13.3)

    def test_pipeline_as_model_input_quickmethod(self):
        """
        Test that visualizer can handle sklearn pipeline as model input
        within a quickmethod
        """
        X, y = load_occupancy(return_dataset=True).to_pandas()

        X_train, X_test, y_train, y_test = tts(
                    X, y, test_size=0.2, shuffle=True, random_state=42
                )

        model = Pipeline([
            ('minmax', MinMaxScaler()), 
            ('svc', SVC(random_state=42))
        ])

        oz = class_prediction_error(model,
                                   X_train, y_train, X_test, y_test,
                                   classes=["vacant", "occupied"],
                                   show=False)
        self.assert_images_similar(oz, tol=12.5, windows_tol=13.3)


# tests.test_classifier.test_classification_report
# Tests for the classification report visualizer
#
# Author:  Rebecca Bilbro
# Author:  Benjamin Bengfort
# Created: Sun Mar 18 16:57:27 2018 -0400
#
# Copyright (C) 208 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_classification_report.py [] benjamin@bengfort.com $

"""
Tests for the classification report visualizer
"""

##########################################################################
## Imports
##########################################################################

import sys
import pytest
import yellowbrick as yb
import matplotlib.pyplot as plt

from yellowbrick.datasets import load_occupancy
from yellowbrick.classifier.classification_report import *

from pytest import approx
from unittest.mock import patch
from tests.base import VisualTestCase

from sklearn.svm import LinearSVC, SVC
from sklearn.naive_bayes import GaussianNB
from sklearn.tree import DecisionTreeClassifier
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split as tts
from sklearn.linear_model import LassoCV, LogisticRegression
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler

try:
    import pandas as pd
except ImportError:
    pd = None


##########################################################################
##  Test for Classification Report
##########################################################################


@pytest.mark.usefixtures("binary", "multiclass")
class TestClassificationReport(VisualTestCase):
    """
    ClassificationReport visualizer tests
    """

    def test_binary_class_report(self):
        """
        Correctly generates a report for binary classification with LinearSVC
        """
        _, ax = plt.subplots()

        viz = ClassificationReport(LinearSVC(random_state=42), ax=ax)
        viz.fit(self.binary.X.train, self.binary.y.train)
        viz.score(self.binary.X.test, self.binary.y.test)

        self.assert_images_similar(viz, tol=40)

        assert viz.scores_ == {
            "precision": {0: approx(0.7446808), 1: approx(0.8490566)},
            "recall": {0: approx(0.8139534), 1: approx(0.7894736)},
            "f1": {0: approx(0.7777777), 1: approx(0.8181818)},
        }

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_multiclass_class_report(self):
        """
        Correctly generates report for multi-class with LogisticRegression
        """
        _, ax = plt.subplots()

        viz = ClassificationReport(LogisticRegression(random_state=12), ax=ax)
        viz.fit(self.multiclass.X.train, self.multiclass.y.train)
        viz.score(self.multiclass.X.test, self.multiclass.y.test)

        self.assert_images_similar(viz, tol=11.0)

        expected = {
            "precision": {
                0: 0.75,
                1: 0.47368421052631576,
                2: 0.45,
                3: 0.375,
                4: 0.5,
                5: 0.5294117647058824,
            },
            "recall": {
                0: 0.47368421052631576,
                1: 0.5625,
                2: 0.6428571428571429,
                3: 0.3157894736842105,
                4: 0.5,
                5: 0.5625,
            },
            "f1": {
                0: 0.5806451612903226,
                1: 0.5142857142857142,
                2: 0.5294117647058824,
                3: 0.34285714285714286,
                4: 0.5,
                5: 0.5454545454545455,
            },
        }
        # Use approx for floating point comparison
        for metric in expected:
            for cls in expected[metric]:
                assert viz.scores_[metric][cls] == approx(expected[metric][cls])

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_pandas_integration(self):
        """
        Test with Pandas DataFrame and Series input
        """
        _, ax = plt.subplots()

        # Load the occupancy dataset from fixtures
        data = load_occupancy(return_dataset=True)
        X, y = data.to_pandas()

        # Create train/test splits
        splits = tts(X, y, test_size=0.2, random_state=4512)
        X_train, X_test, y_train, y_test = splits

        classes = ["unoccupied", "occupied"]

        # Create classification report
        model = GaussianNB()
        viz = ClassificationReport(model, ax=ax, classes=classes)
        viz.fit(X_train, y_train)
        viz.score(X_test, y_test)

        self.assert_images_similar(viz, tol=5.0)

        # Ensure correct classification scores under the hood!
        assert viz.scores_ == {
            "precision": {
                "unoccupied": 0.999347471451876,
                "occupied": 0.8825214899713467,
            },
            "recall": {
                "unoccupied": 0.9613935969868174,
                "occupied": 0.9978401727861771,
            },
            "f1": {"unoccupied": 0.9800031994880819, "occupied": 0.9366447034972124},
        }

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_numpy_integration(self):
        """
        Test with NumPy arrays
        """
        _, ax = plt.subplots()

        # Load the occupancy dataset from fixtures
        data = load_occupancy(return_dataset=True)
        X, y = data.to_numpy()

        # Create train/test splits
        splits = tts(X, y, test_size=0.2, random_state=4512)
        X_train, X_test, y_train, y_test = splits

        classes = ["unoccupied", "occupied"]

        # Create classification report
        model = GaussianNB()
        viz = ClassificationReport(model, ax=ax, classes=classes)
        viz.fit(X_train, y_train)
        viz.score(X_test, y_test)

        self.assert_images_similar(viz, tol=5.0)

        # Ensure correct classification scores under the hood!
        assert viz.scores_ == {
            "precision": {
                "unoccupied": 0.999347471451876,
                "occupied": 0.8825214899713467,
            },
            "recall": {
                "unoccupied": 0.9613935969868174,
                "occupied": 0.9978401727861771,
            },
            "f1": {"unoccupied": 0.9800031994880819, "occupied": 0.9366447034972124},
        }

    def test_quick_method(self):
        """
        Test the quick method with a random dataset
        """
        X, y = make_classification(
            n_samples=400,
            n_features=20,
            n_informative=8,
            n_redundant=8,
            n_classes=2,
            n_clusters_per_class=4,
            random_state=27,
        )

        # Create train/test splits
        splits = tts(X, y, test_size=0.2, random_state=42)
        X_train, X_test, y_train, y_test = splits

        _, ax = plt.subplots()
        model = DecisionTreeClassifier(random_state=19)
        visualizer = classification_report(
            model, X_train, y_train, X_test, y_test, ax=ax, show=False
        )

        assert isinstance(visualizer, ClassificationReport)
        self.assert_images_similar(visualizer, tol=12)

    def test_isclassifier(self):
        """
        Assert that only classifiers can be used with the visualizer.
        """

        message = (
            "This estimator is not a classifier; "
            "try a regression or clustering score visualizer instead!"
        )

        with pytest.raises(yb.exceptions.YellowbrickError, match=message):
            ClassificationReport(LassoCV())

    def test_support_count_class_report(self):
        """
        Correctly generates a report showing support as a raw count
        """
        _, ax = plt.subplots()

        viz = ClassificationReport(LinearSVC(random_state=42), ax=ax, support="count")
        viz.fit(self.binary.X.train, self.binary.y.train)
        viz.score(self.binary.X.test, self.binary.y.test)

        self.assert_images_similar(viz, tol=40)

        assert viz.scores_ == {
            "precision": {0: approx(0.7446808), 1: approx(0.8490566)},
            "recall": {0: approx(0.8139534), 1: approx(0.7894736)},
            "f1": {0: approx(0.7777777), 1: approx(0.8181818)},
            "support": {0: approx(0.42999999999999999), 1: approx(0.56999999999999995)},
        }

    def test_support_percent_class_report(self):
        """
        Correctly generates a report showing support as a percent
        """
        _, ax = plt.subplots()

        viz = ClassificationReport(LinearSVC(random_state=42), ax=ax, support="percent")
        viz.fit(self.binary.X.train, self.binary.y.train)
        viz.score(self.binary.X.test, self.binary.y.test)

        self.assert_images_similar(viz, tol=40)

        assert viz.scores_ == {
            "precision": {0: approx(0.7446808), 1: approx(0.8490566)},
            "recall": {0: approx(0.8139534), 1: approx(0.7894736)},
            "f1": {0: approx(0.7777777), 1: approx(0.8181818)},
            "support": {0: approx(0.42999999999999999), 1: approx(0.56999999999999995)},
        }

    def test_invalid_support(self):
        """
        Ensure that bad support arguments raise exception
        """
        with pytest.raises(
            YellowbrickValueError,
            match="'foo' is an invalid argument for support, use None, "
            "True, False, 'percent', or 'count'",
        ):
            ClassificationReport(LinearSVC(), support="foo")

    def test_score_returns_score(self):
        """
        Test that ClassificationReport score() returns a score between 0 and 1
        """
        viz = ClassificationReport(LinearSVC(random_state=42))

        viz.fit(self.binary.X.train, self.binary.y.train)
        s = viz.score(self.binary.X.test, self.binary.y.test)

        assert 0 <= s <= 1

    def test_with_fitted(self):
        """
        Test that visualizer properly handles an already-fitted model
        """
        X, y = load_occupancy(return_dataset=True).to_numpy()

        model = LinearSVC().fit(X, y)
        classes = ["unoccupied", "occupied"]

        with patch.object(model, "fit") as mockfit:
            oz = ClassificationReport(model, classes=classes)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with patch.object(model, "fit") as mockfit:
            oz = ClassificationReport(model, classes=classes, is_fitted=True)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with patch.object(model, "fit") as mockfit:
            oz = ClassificationReport(model, classes=classes, is_fitted=False)
            oz.fit(X, y)
            mockfit.assert_called_once_with(X, y)

    def test_remove_color_bar(self):
        """
        Correctly removes the colorbar for binary classification with LinearSVC
        """
        _, ax = plt.subplots()

        viz = ClassificationReport(LinearSVC(random_state=42), ax=ax, colorbar=False)
        viz.fit(self.binary.X.train, self.binary.y.train)
        viz.score(self.binary.X.test, self.binary.y.test)

        self.assert_images_similar(viz, tol=40)

    def test_with_missing_labels(self):
        """
        Test that visualizer properly handles missing labels when scoring
        """
        _, ax = plt.subplots()

        X_train = np.array([[1], [2], [3]])
        y_train = np.array([0, 1, 2])

        X_test = np.array([[1], [2]])
        y_test = np.array([0, 1])

        viz = ClassificationReport(LogisticRegression(), ax=ax)
        viz.fit(X_train, y_train)
        viz.score(X_test, y_test)

        assert viz.scores_ == {
            "precision": {0: approx(1.0), 1: approx(1.0), 2: approx(0.0)},
            "recall": {0: approx(1.0), 1: approx(1.0), 2: approx(0.0)},
            "f1": {0: approx(1.0), 1: approx(1.0), 2: approx(0.0)},
        }

    def test_within_pipeline(self):
        """
        Test that visualizer can be accessed within a sklearn pipeline
        """
        X, y = load_occupancy(return_dataset=True).to_pandas()
        classes = ["unoccupied", "occupied"]

        X_train, X_test, y_train, y_test = tts(
            X, y, test_size=0.2, shuffle=True, random_state=42
        )

        model = Pipeline(
            [
                ("minmax", MinMaxScaler()),
                ("clsrpt", ClassificationReport(SVC(random_state=42), classes=classes)),
            ]
        )

        model.fit(X_train, y_train)
        model.score(X_test, y_test)
        model["clsrpt"].finalize()
        self.assert_images_similar(model["clsrpt"], tol=15)

    def test_within_pipeline_quickmethod(self):
        """
        Test that visualizer quickmethod can be accessed within a
        sklearn pipeline
        """
        X, y = load_occupancy(return_dataset=True).to_pandas()

        X_train, X_test, y_train, y_test = tts(
            X, y, test_size=0.2, shuffle=True, random_state=42
        )

        model = Pipeline(
            [
                ("minmax", MinMaxScaler()),
                (
                    "clsrpt",
                    classification_report(
                        SVC(random_state=42),
                        X_train,
                        y_train,
                        X_test,
                        y_test,
                        classes=["vacant", "occupied"],
                        show=False,
                    ),
                ),
            ]
        )
        self.assert_images_similar(model["clsrpt"], tol=15)

    def test_pipeline_as_model_input(self):
        """
        Test that visualizer can handle sklearn pipeline as model input
        """
        X, y = load_occupancy(return_dataset=True).to_pandas()
        classes = ["unoccupied", "occupied"]

        X_train, X_test, y_train, y_test = tts(
            X, y, test_size=0.2, shuffle=True, random_state=42
        )

        model = Pipeline([("minmax", MinMaxScaler()), ("svc", SVC(random_state=42))])

        oz = ClassificationReport(model, classes=classes)
        oz.fit(X_train, y_train)
        oz.score(X_test, y_test)
        oz.finalize()
        self.assert_images_similar(oz, tol=15)

    def test_pipeline_as_model_input_quickmethod(self):
        """
        Test that visualizer can handle sklearn pipeline as model input
        within a quickmethod
        """
        X, y = load_occupancy(return_dataset=True).to_pandas()

        X_train, X_test, y_train, y_test = tts(
            X, y, test_size=0.2, shuffle=True, random_state=42
        )

        model = Pipeline([("minmax", MinMaxScaler()), ("svc", SVC(random_state=42))])

        oz = classification_report(
            model,
            X_train,
            y_train,
            X_test,
            y_test,
            classes=["vacant", "occupied"],
            show=False,
        )
        self.assert_images_similar(oz, tol=15)



# tests.base
# Helper functions and cases for making assertions on visualizations.
#
# Author:   Benjamin Bengfort <bbengfort@districtdatalabs.com>
# Created:  Sun Oct 09 12:23:13 2016 -0400
#
# ID: base.py [b8e3318] benjamin@bengfort.com $

"""
Helper functions and cases for making assertions on visualizations.
"""

##########################################################################
## Imports
##########################################################################

import os
import sys
import inspect

import matplotlib as mpl
import matplotlib.pyplot as plt

from matplotlib import ticker
from matplotlib.testing.compare import compare_images
from yellowbrick.exceptions import ImageComparisonFailure


##########################################################################
## Environment
##########################################################################


def is_windows_or_conda():
    """
    Simple detection mechanism to determine if the tests are running in a
    win32 or Anaconda/Miniconda environment.
    """
    is_windows = sys.platform == "win32"
    is_conda = os.path.exists(os.path.join(sys.prefix, "conda-meta"))
    return is_windows or is_conda


##########################################################################
## Module Constants
##########################################################################

# The root directory for all tests
TESTS = os.path.dirname(__file__)
ACTUAL_IMAGES = os.path.join(TESTS, "actual_images")
BASELINE_IMAGES = os.path.join(TESTS, "baseline_images")
IS_WINDOWS_OR_CONDA = is_windows_or_conda()


##########################################################################
## Visual Test Case
##########################################################################


class VisualTestCase(object):
    """
    The visual test case class ensures that all tests inside of the class
    can execute image similarity tests inside of a clean matplotlib global
    figure.
    """

    def setup_method(self):
        """
        Before a visual test case method is run, ensure that the previous
        figure is closed and the current axes are cleared.

        See: https://docs.pytest.org/en/latest/xunit_setup.html
        """
        # Reset the matplotlib environment
        plt.cla()  # clear current axis
        plt.clf()  # clear current figure
        plt.close("all")  # close all existing plots

        # Travis-CI does not have san-serif so ensure standard fonts are used.
        # Note that this must be set before each test otherwise it will be reset by
        # the Yellowbrick styles.
        mpl.rcParams["font.family"] = "DejaVu Sans"

    def assert_images_similar(
        self, visualizer=None, ax=None, tol=0.01, windows_tol=None, **kwargs
    ):
        """Accessible testing method for testing generation of a Visualizer.

        Requires the placement of a baseline image for comparison in the
        tests/baseline_images folder that corresponds to the module path of the
        VisualTestCase being evaluated. The name of the image corresponds to
        the test function where "self.assert_images_similar" is called.

        For example, calling "assert_images_similar" in the test function
        "test_class_report" in tests.test_classifier.test_class_balance would
        require placement a baseline image at:

        baseline_images/test_classifier/test_class_balance/test_class_report.png

        The easiest way to generate a baseline image is to first run the test that
        calls "assert_images_similar", and then copy the actual test generated
        image from:

        actual_images/

        visualizer : yellowbrick visualizer, default: None
            An instantiated yellowbrick visualizer that has been fitted,
            transformed and had all operations except for show called on it.

        ax : matplotlib Axes, default: None
            The axis to plot the figure on.

        tol : float, default: 0.01
            The tolerance (a color value difference, where 255 is the
            maximal difference).  The test fails if the average pixel
            difference is greater than this value.

        windows_tol: float, default: None
            Similar to the tol parameter, but targeted for testing on a
            windows environment.

        kwargs : dict
            Options to pass to the ImageComparison class.
        """
        # Hide this method from the pytest traceback on test failure.
        __tracebackhide__ = True

        # Build and execute the image comparison
        compare = ImageComparison(
            inspect.stack(),
            visualizer=visualizer,
            ax=ax,
            tol=tol,
            windows_tol=windows_tol,
            **kwargs
        )
        compare()

        # Return the compare object for meta testing
        return compare


##########################################################################
## Image Comparison Test
##########################################################################


class ImageComparison(object):
    """
    An image comparison wraps a single ``assert_images_similar`` statement to
    compose the actual and baseline image paths based on the stack the
    assertion was called in. It contains all properties that were formerly
    set on the test case (to facilitate our transition to pytest) so that they
    are immutable with respect to the single image comparison. By
    encapsulating these details, it is easier to debug image comparisons in
    meta tests.

    Parameters
    ----------
    stack : list of FrameInfo
        The list of the frame records for the caller's stack obtained from the
        ``inspect.stack()`` function. Must be called from the entry point to
        the image comparison (e.g. the next function from the test function).

    visualizer : Yellowbrick Visualizer instance, optional
        An instantiated Yellowbrick visualizer that wraps a matplotlib Axes
        and has been drawn on via the Yellowbrick API. If a visualizer is specified,
        generally speaking it should have a call to ``finalize()`` performed, but not a
        call to ``show()``.

    ax : matplotlib Axes, optional
        A direct reference to a matplotlib Axes that has been drawn on. If using the
        Axes object, you should not call ``plt.show()``.

    tol : float, default: 0.01
        The tolerance as a color value difference, where 255 is the maximal
        difference. The test fails if the average pixel difference is greater
        than this value.

    windows_tol : float, default: 0.01
        The tolerance (tol) parameter for the windows operating system environment.

    ext : string, default: ".png"
        The file extension to save the actual and baseline images as.

    remove_ticks : bool, default: True
        Remove the major and minor ticks on both the y and x axis to make the
        image comparison simpler (since different OS may have varying fonts or
        system level preferences).

    remove_title : bool, default: True
        Remove the title since different OS may have varying fonts.

    remove_labels : bool, default: True
        Remove the x and y labels since different OS may have varying fonts.

    remove_legend : bool, default: True
        Remove the legend since different OS may have varying fonts.

    Raises
    ------
    ValueError : at least one of visualizer or ax must be specified.
    """

    def __init__(
        self,
        stack,
        visualizer=None,
        ax=None,
        tol=0.01,
        windows_tol=0.01,
        ext=".png",
        remove_ticks=True,
        remove_title=True,
        remove_labels=True,
        remove_legend=True,
    ):

        # Ensure we have something to draw on
        if visualizer is None and ax is None:
            raise ValueError("at least one of visualizer or ax must be specified")

        # Save the ax being drawn on
        self.ax = ax or visualizer.ax

        # Parse the stack for the test and module name, element 0 should be
        # the assertion function (or whatever called this init), element 1
        # should be the test function and start with the test.
        frame = stack[1]

        # FrameInfo(frame, filename, lineno, function, code_context, index)
        self.test_func_name = frame[3]
        if not self.test_func_name.startswith("test"):
            raise ValueError("{} is not a test function".format(self.test_func_name))

        # Find the relative path to the Yellowbrick tests to compute the
        # module name for storing images in the actual and baseline dirs.
        root = os.path.commonprefix((TESTS, frame[1]))
        module_path = os.path.relpath(frame[1], root)
        self.test_module_path = os.path.splitext(module_path)[0]

        # Set the error tolerance depending on the os
        if os.name == "nt" and windows_tol is not None:
            self.tol = windows_tol
        else:
            self.tol = tol

        # Save other image comparison properties
        self.ext = ext
        self.remove_ticks = remove_ticks
        self.remove_title = remove_title
        self.remove_labels = remove_labels
        self.remove_legend = remove_legend

    def __call__(self):
        """
        Executes the image comparison by cleaning up the actual figure, saving
        the actual figure, then comparing the actual figure to the baseline.
        """
        # Hide this method from the pytest traceback on test failure.
        __tracebackhide__ = True

        self.cleanup()
        self.save()
        self.compare()

    @property
    def actual_image_path(self):
        """
        Computes the path in ACTUAL_IMAGES to the test image based on the test
        name and module. Creates any required parent dirs along the way.
        """
        return self._image_path(ACTUAL_IMAGES)

    @property
    def baseline_image_path(self):
        """
        Computes the path in BASELINE_IMAGES to the expected location of the
        file to compare the actual image against. Creates any required parent
        dirs along the way.
        """
        return self._image_path(BASELINE_IMAGES)

    def _image_path(self, root):
        """
        Computes the image path from the specified root directory (shared
        functionality for both actual and baseline image paths).
        """
        # Directory the images for this module are stored in.
        imgdir = os.path.join(root, self.test_module_path)

        # Create directory if it doesn't exist
        if not os.path.exists(imgdir):
            os.makedirs(imgdir, mode=0o777, exist_ok=True)

        # Create the image path from the test name
        return os.path.join(imgdir, self.test_func_name + self.ext)

    def cleanup(self):
        """
        Cleanup the image by removing textual/formatting elements.
        """
        # Hide this method from the pytest traceback on test failure.
        __tracebackhide__ = True

        if self.remove_title:
            self.ax.set_title("")

        if self.remove_ticks:
            null_formatter = ticker.NullFormatter()
            for axis in ("xaxis", "yaxis", "zaxis"):
                try:
                    axis = getattr(self.ax, axis)
                    axis.set_major_formatter(null_formatter)
                    axis.set_minor_formatter(null_formatter)
                except AttributeError:
                    continue

        if self.remove_labels:
            self.ax.set_xlabel("")
            self.ax.set_ylabel("")

        if self.remove_legend and self.ax.get_legend() is not None:
            self.ax.legend_.remove()

    def save(self):
        """
        Save the actual image to disk after cleaning it up.
        """
        # Hide this method from the pytest traceback on test failure.
        __tracebackhide__ = True

        assert self.ax.has_data(), "nothing has been drawn on the Axes"
        self.ax.get_figure().savefig(self.actual_image_path)

    def compare(self):
        """
        Compare the actual image to the baseline image, raising an exception
        if the baseline image does not exist or if the image is not close.
        """
        # Hide this method from the pytest traceback on test failure.
        __tracebackhide__ = True

        # Get expected and actual paths
        expected = self.baseline_image_path
        actual = self.actual_image_path

        # Assert that we have an actual image already saved
        assert os.path.exists(actual), "actual image hasn't been saved yet"

        # Ensure we have an image to compare against (common failure)
        if not os.path.exists(expected):
            raise ImageComparisonFailure(
                "baseline image does not exist:\n{}".format(os.path.relpath(expected))
            )
        # Perform the comparison
        err = compare_images(expected, actual, self.tol, in_decorator=True)

        # Raise image comparison failure if not close
        if err:
            for key in ("actual", "expected"):
                err[key] = os.path.relpath(err[key])

            raise ImageComparisonFailure(
                (
                    "images not close (RMS {rms:0.3f})" "\n{actual}\n\tvs\n{expected}"
                ).format(**err)
            )



# tests.test_features.test_radviz
# Test the RadViz feature analysis visualizers
#
# Author:   Benjamin Bengfort
# Created:  Fri Oct 07 12:19:19 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_radviz.py [01d5996] benjamin@bengfort.com $

"""
Test the RadViz feature analysis visualizers
"""

##########################################################################
## Imports
##########################################################################

import sys
import pytest
import numpy.testing as npt

from tests.base import IS_WINDOWS_OR_CONDA, VisualTestCase
from ..fixtures import Dataset
from sklearn.datasets import make_classification

from yellowbrick.datasets import load_occupancy
from yellowbrick.features.radviz import *

try:
    import pandas as pd
except ImportError:
    pd = None

##########################################################################
## Fixtures
##########################################################################


@pytest.fixture(scope="class")
def dataset(request):
    """
    Creates a random multiclass classification dataset fixture
    """
    X, y = make_classification(
        n_samples=200,
        n_features=5,
        n_informative=4,
        n_redundant=0,
        n_classes=3,
        n_clusters_per_class=1,
        random_state=451,
        flip_y=0,
        class_sep=3,
        scale=np.array([1.0, 2.0, 100.0, 20.0, 1.0]),
    )

    dataset = Dataset(X, y)
    request.cls.dataset = dataset


##########################################################################
## RadViz Tests
##########################################################################


@pytest.mark.usefixtures("dataset")
class TestRadViz(VisualTestCase):
    """
    Test the RadViz visualizer
    """

    def test_normalize_x(self):
        """
        Test the static normalization method on the RadViz class
        """
        # Original data
        X = np.array(
            [
                [2.318, 2.727, 4.260, 7.212, 4.792],
                [2.315, 2.726, 4.295, 7.140, 4.783],
                [2.315, 2.724, 4.260, 7.135, 4.779],
                [2.110, 3.609, 4.330, 7.985, 5.595],
                [2.110, 3.626, 4.330, 8.203, 5.621],
                [2.110, 3.620, 4.470, 8.210, 5.612],
            ]
        )

        # Expected result
        Xe = np.array(
            [
                [1.0, 0.00332594, 0.0, 0.07162791, 0.01543943],
                [0.98557692, 0.00221729, 0.16666667, 0.00465116, 0.00475059],
                [0.98557692, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.98115299, 0.33333333, 0.79069767, 0.96912114],
                [0.0, 1.0, 0.33333333, 0.99348837, 1.0],
                [0.0, 0.99334812, 1.0, 1.0, 0.98931116],
            ]
        )

        # Xprime (transformed X)
        Xp = RadViz.normalize(X)
        npt.assert_array_almost_equal(Xp, Xe)

    def test_radviz(self):
        """
        Assert image similarity on test dataset
        """
        visualizer = RadViz()
        visualizer.fit_transform(self.dataset.X, self.dataset.y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.25)

    def test_radviz_alpha(self):
        """
        Assert image similarity with alpha transparency
        """
        visualizer = RadViz(alpha=0.5)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.25)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_integrated_radviz_with_pandas(self):
        """
        Test RadViz with Pandas on the occupancy dataset
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_pandas()

        assert isinstance(X, pd.DataFrame)
        assert isinstance(y, pd.Series)

        # Test the visualizer
        visualizer = RadViz()
        visualizer.fit_transform(X, y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.1)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_integrated_radviz_with_numpy(self):
        """
        Test RadViz with numpy on the occupancy dataset
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_numpy()

        assert isinstance(X, np.ndarray)
        assert isinstance(y, np.ndarray)

        # Test the visualizer
        visualizer = RadViz()
        visualizer.fit_transform(X, y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.1)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_integrated_radviz_pandas_classes_features(self):
        """
        Test RadViz with classes and features specified using Pandas
        """
        # Load the data from the fixture
        data = load_occupancy(return_dataset=True)
        X, y = data.to_pandas()

        features = ["temperature", "relative humidity", "light"]
        classes = [
            k for k, _ in sorted(data.meta["labels"].items(), key=lambda i: i[1])
        ]

        assert isinstance(X, pd.DataFrame)
        assert isinstance(y, pd.Series)

        # Filter the dataset to make sure it's not just class names
        X = X[features]
        y = y.astype(int)

        # Test the visualizer
        visualizer = RadViz(features=features, classes=classes)
        visualizer.fit_transform(X, y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.1)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_integrated_radviz_numpy_classes_features(self):
        """
        Test RadViz with classes and features specified using numpy
        """
        # Load the data from the fixture
        data = load_occupancy(return_dataset=True)
        X, y = data.to_numpy()

        features = data.meta["features"][0:3]
        classes = [
            k for k, _ in sorted(data.meta["labels"].items(), key=lambda i: i[1])
        ]

        assert isinstance(X, np.ndarray)
        assert isinstance(y, np.ndarray)

        # Filter the dataset to make sure it's not just class names
        X = X[:, :3]
        y = y.astype(int)

        # Test the visualizer
        visualizer = RadViz(features=features, classes=classes)
        visualizer.fit_transform(X, y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.1)

    def test_radviz_quick_method(self):
        """
        Test RadViz quick method with colors being set.
        """
        visualizer = radviz(
            *self.dataset, colors=["cyan", "magenta", "yellow"], show=False
        )
        self.assert_images_similar(visualizer)



# tests.test_datasets.test_download
# Tests the download from S3 to ensure data is accessible.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Tue Jan 01 15:06:05 2019 -0500
#
# For license information, see LICENSE.txt
#
# ID: test_download.py [57aab02] ndanielsen@users.noreply.github.com $

"""
Tests the download from S3 to ensure data is accessible.
"""

##########################################################################
## Imports
##########################################################################

import pytest

from yellowbrick.datasets.loaders import *
from yellowbrick.datasets.loaders import DATASETS
from yellowbrick.datasets.path import dataset_exists, dataset_archive


@pytest.mark.parametrize(
    "loader",
    [
        load_bikeshare,
        load_concrete,
        load_credit,
        load_energy,
        load_game,
        load_mushroom,
        load_occupancy,
        load_spam,
        load_walking,
        load_hobbies,
        load_nfl,
    ],
    ids=lambda l: l.__name__,
)
def test_loader_download(tmpdir, loader):
    """
    Test download of dataset when it does not exist (requires Internet connection!)
    """
    name = loader.__name__[len("load_") :]
    data_home = str(tmpdir.mkdir("datasets"))

    # The dataset should not exist
    assert not dataset_exists(name, data_home=data_home)
    assert not dataset_archive(name, DATASETS[name]["signature"], data_home=data_home)

    # Load the dataset
    loader(data_home=data_home)

    # The dataset should have been downloaded
    assert dataset_exists(name, data_home=data_home)
    assert dataset_archive(name, DATASETS[name]["signature"], data_home=data_home)



# tests.test_cluster.test_silhouette
# Tests for the SilhouetteVisualizer
#
# Author:   Benjamin Bengfort
# Created:  Mon Mar 27 10:01:37 2017 -0400
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_silhouette.py [57b563b] benjamin@bengfort.com $

"""
Tests for the SilhouetteVisualizer
"""

##########################################################################
## Imports
##########################################################################

import sys
import pytest
import matplotlib.pyplot as plt
import numpy as np

from sklearn.datasets import make_blobs
from sklearn.cluster import KMeans, MiniBatchKMeans
from sklearn.cluster import SpectralClustering, AgglomerativeClustering

from unittest import mock
from tests.base import VisualTestCase

from yellowbrick.cluster.silhouette import SilhouetteVisualizer, silhouette_visualizer


##########################################################################
## SilhouetteVisualizer Test Cases
##########################################################################


class TestSilhouetteVisualizer(VisualTestCase):
    """
    Silhouette Visualizer Tests
    """

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_integrated_kmeans_silhouette(self):
        """
        Test no exceptions for kmeans silhouette visualizer on blobs dataset
        """
        # NOTE see #182: cannot use occupancy dataset because of memory usage

        # Generate a blobs data set
        X, y = make_blobs(
            n_samples=1000, n_features=12, centers=8, shuffle=False, random_state=0
        )

        fig = plt.figure()
        ax = fig.add_subplot()

        visualizer = SilhouetteVisualizer(KMeans(random_state=0), ax=ax)
        visualizer.fit(X)
        visualizer.finalize()

        self.assert_images_similar(visualizer, remove_legend=True)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_integrated_mini_batch_kmeans_silhouette(self):
        """
        Test no exceptions for mini-batch kmeans silhouette visualizer
        """
        # NOTE see #182: cannot use occupancy dataset because of memory usage

        # Generate a blobs data set
        X, y = make_blobs(
            n_samples=1000, n_features=12, centers=8, shuffle=False, random_state=0
        )

        fig = plt.figure()
        ax = fig.add_subplot()

        visualizer = SilhouetteVisualizer(MiniBatchKMeans(random_state=0), ax=ax)
        visualizer.fit(X)
        visualizer.finalize()

        self.assert_images_similar(visualizer, remove_legend=True)

    @pytest.mark.skip(reason="no negative silhouette example available yet")
    def test_negative_silhouette_score(self):
        """
        Ensure negative silhouette scores are correctly displayed by the visualizer.
        """
        raise NotImplementedError("no negative silhouette example available")

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_colormap_silhouette(self):
        """
        Test no exceptions for modifying the colormap in a silhouette visualizer
        """
        # Generate a blobs data set
        X, y = make_blobs(
            n_samples=1000, n_features=12, centers=8, shuffle=False, random_state=0
        )

        fig = plt.figure()
        ax = fig.add_subplot()

        visualizer = SilhouetteVisualizer(
            MiniBatchKMeans(random_state=0), ax=ax, colormap="gnuplot"
        )
        visualizer.fit(X)
        visualizer.finalize()

        self.assert_images_similar(visualizer, remove_legend=True)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_colors_silhouette(self):
        """
        Test no exceptions for modifying the colors in a silhouette visualizer
        with a list of color names
        """
        # Generate a blobs data set
        X, y = make_blobs(
            n_samples=1000, n_features=12, centers=8, shuffle=False, random_state=0
        )

        fig = plt.figure()
        ax = fig.add_subplot()

        visualizer = SilhouetteVisualizer(
            MiniBatchKMeans(random_state=0),
            ax=ax,
            colors=["red", "green", "blue", "indigo", "cyan", "lavender"],
        )
        visualizer.fit(X)
        visualizer.finalize()

        self.assert_images_similar(visualizer, remove_legend=True)

    def test_colormap_as_colors_silhouette(self):
        """
        Test no exceptions for modifying the colors in a silhouette visualizer
        by using a matplotlib colormap as colors
        """
        # Generate a blobs data set
        X, y = make_blobs(
            n_samples=1000, n_features=12, centers=8, shuffle=False, random_state=0
        )

        fig = plt.figure()
        ax = fig.add_subplot()

        visualizer = SilhouetteVisualizer(
            MiniBatchKMeans(random_state=0), ax=ax, colors="cool"
        )
        visualizer.fit(X)
        visualizer.finalize()

        tol = (
            3.2 if sys.platform == "win32" else 0.01
        )  # Fails on AppVeyor with RMS 3.143
        self.assert_images_similar(visualizer, remove_legend=True, tol=tol)

    def test_quick_method(self):
        """
        Test the quick method producing a valid visualization
        """
        X, y = make_blobs(
            n_samples=1000, n_features=12, centers=8, shuffle=False, random_state=0
        )

        model = MiniBatchKMeans(3, random_state=343)
        oz = silhouette_visualizer(model, X, show=False)
        assert isinstance(oz, SilhouetteVisualizer)

        self.assert_images_similar(oz)

    def test_with_fitted(self):
        """
        Test that visualizer properly handles an already-fitted model
        """
        X, y = make_blobs(
            n_samples=100, n_features=5, centers=3, shuffle=False, random_state=112
        )
        model = MiniBatchKMeans().fit(X)
        labels = model.predict(X)

        with mock.patch.object(model, "fit") as mockfit:
            oz = SilhouetteVisualizer(model)
            oz.fit(X)
            mockfit.assert_not_called()

        with mock.patch.object(model, "fit") as mockfit:
            oz = SilhouetteVisualizer(model, is_fitted=True)
            oz.fit(X)
            mockfit.assert_not_called()

        with mock.patch.object(model, "fit_predict", return_value=labels) as mockfit:
            oz = SilhouetteVisualizer(model, is_fitted=False)
            oz.fit(X)
            mockfit.assert_called_once_with(X, None)

    @pytest.mark.parametrize(
        "model",
        [SpectralClustering, AgglomerativeClustering],
    )
    def test_clusterer_without_predict(self, model):
        """
        Assert that clustering estimators that don't implement
        a predict() method utilize fit_predict()
        """
        X = np.array([[1, 2], [1, 4], [1, 0], [4, 2], [4, 4], [4, 0]])
        try:
            visualizer = SilhouetteVisualizer(model(n_clusters=2))
            visualizer.fit(X)
            visualizer.finalize()
        except AttributeError:
            self.fail("could not use fit or fit_predict methods")



# tests.test_cluster.test_icdm
# Tests for the intercluster distance map visualizer.
#
# Author:  Benjamin Bengfort
# Created: Tue Aug 21 11:57:44 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_icdm.py [2f23976] benjamin@bengfort.com $

"""
Tests for the intercluster distance map visualizer.
"""

##########################################################################
## Imports
##########################################################################

import pytest
import matplotlib as mpl
import numpy as np

from yellowbrick.cluster.icdm import *
from yellowbrick.datasets import load_nfl
from yellowbrick.exceptions import YellowbrickValueError

from unittest import mock
from tests.fixtures import Dataset
from tests.base import IS_WINDOWS_OR_CONDA, VisualTestCase

from sklearn.datasets import make_blobs
from sklearn.cluster import Birch, AgglomerativeClustering
from sklearn.decomposition import LatentDirichletAllocation as LDA
from sklearn.cluster import KMeans, AffinityPropagation, MiniBatchKMeans
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer

try:
    import pandas as pd
except ImportError:
    pd = None

# Determine version of matplotlib
MPL_VERS_MAJ = int(mpl.__version__.split(".")[0])

##########################################################################
## Fixtures
##########################################################################


@pytest.fixture(scope="class")
def blobs12(request):
    """
    Creates a fixture of 1000 instances in 12 clusters with 16 features.
    """
    X, y = make_blobs(
        centers=12, n_samples=1000, n_features=16, shuffle=True, random_state=2121
    )
    request.cls.blobs12 = Dataset(X, y)


@pytest.fixture(scope="class")
def blobs4(request):
    """
    Creates a fixture of 400 instances in 4 clusters with 16 features.
    """
    X, y = make_blobs(
        centers=4, n_samples=400, n_features=16, shuffle=True, random_state=1212
    )
    request.cls.blobs4 = Dataset(X, y)


def assert_fitted(oz):
    for param in ("cluster_centers_", "embedded_centers_", "scores_", "fit_time_"):
        assert hasattr(oz, param)


def assert_not_fitted(oz):
    for param in ("embedded_centers_", "scores_", "fit_time_"):
        assert not hasattr(oz, param)


##########################################################################
## InterclusterDistance Test Cases
##########################################################################


@pytest.mark.usefixtures("blobs12", "blobs4")
class TestInterclusterDistance(VisualTestCase):
    """
    Test the InterclusterDistance visualizer
    """

    def test_only_valid_embeddings(self):
        """
        Should raise an exception on invalid embedding
        """
        # On init
        with pytest.raises(YellowbrickValueError, match="unknown embedding 'foo'"):
            InterclusterDistance(KMeans(), embedding="foo")

        # After init
        icdm = InterclusterDistance(KMeans())
        icdm.embedding = "foo"
        with pytest.raises(YellowbrickValueError, match="unknown embedding 'foo'"):
            icdm.transformer

    def test_only_valid_scoring(self):
        """
        Should raise an exception on invalid scoring
        """
        # On init
        with pytest.raises(YellowbrickValueError, match="unknown scoring 'foo'"):
            InterclusterDistance(KMeans(), scoring="foo")

        # After init
        icdm = InterclusterDistance(KMeans())
        icdm.scoring = "foo"
        with pytest.raises(YellowbrickValueError, match="unknown scoring method 'foo'"):
            icdm._score_clusters(None)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_kmeans_mds(self):
        """
        Visual similarity with KMeans and MDS scaling
        """
        model = KMeans(9, random_state=38)
        oz = InterclusterDistance(model, random_state=83, embedding="mds")

        # Prefit assertions
        assert_not_fitted(oz)

        assert oz.fit(self.blobs12.X) is oz  # Fit returns self

        # Postfit assertions
        assert_fitted(oz)
        assert oz.embedded_centers_.shape[0] == oz.scores_.shape[0]
        assert oz.embedded_centers_.shape[0] == oz.cluster_centers_.shape[0]
        assert len(oz._score_clusters(self.blobs12.X)) == 9
        assert len(oz._get_cluster_sizes()) == 9

        # Image similarity
        oz.finalize()
        self.assert_images_similar(oz, tol=0.5)

    @pytest.mark.filterwarnings("ignore:the matrix subclass is not the recommended way")
    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_affinity_tsne_no_legend(self):
        """
        Visual similarity with AffinityPropagation, TSNE scaling, and no legend
        """
        model = AffinityPropagation(random_state=42)
        oz = InterclusterDistance(
            model, random_state=763, embedding="tsne", legend=False, perplexity=3
        )

        # Prefit assertions
        assert_not_fitted(oz)

        assert oz.fit(self.blobs4.X) is oz  # Fit returns self

        # Postfit assertions
        assert_fitted(oz)
        assert oz.embedded_centers_.shape[0] == oz.scores_.shape[0]
        assert oz.embedded_centers_.shape[0] == oz.cluster_centers_.shape[0]

        # Image similarity
        oz.finalize()
        # t-SNE is stochastic, allow for variance across platforms (RMS 9.630 * 1.20)
        self.assert_images_similar(oz, tol=11.6)

    @pytest.mark.skip(reason="LDA not implemented yet")
    def test_lda_mds(self):
        """
        Visual similarity with LDA and MDS scaling
        """
        model = LDA(9, random_state=6667)
        oz = InterclusterDistance(model, random_state=2332, embedding="mds")

        # Prefit assertions
        assert_not_fitted(oz)

        assert oz.fit(self.blobs12.X) is oz  # Fit returns self

        # Postfit assertions
        assert_fitted(oz)
        assert oz.embedded_centers_.shape[0] == oz.scores_.shape[0]
        assert oz.embedded_centers_.shape[0] == oz.cluster_centers_.shape[0]
        assert len(oz._score_clusters(self.blobs12.X)) == 9
        assert len(oz._get_cluster_sizes()) == 9

        # Image similarity
        oz.finalize()
        self.assert_images_similar(oz, tol=1.0)

    @pytest.mark.skip(reason="agglomerative not implemented yet")
    @pytest.mark.filterwarnings("ignore:Using a non-tuple sequence")
    @pytest.mark.filterwarnings("ignore:the matrix subclass is not the recommended way")
    def test_birch_tsne(self):
        """
        Visual similarity with Birch and MDS scaling
        """
        oz = InterclusterDistance(Birch(n_clusters=9), random_state=83, embedding="mds")

        # Prefit assertions
        assert_not_fitted(oz)

        assert oz.fit(self.blobs12.X) is oz  # Fit returns self

        # Postfit assertions
        assert_fitted(oz)
        assert oz.embedded_centers_.shape[0] == oz.scores_.shape[0]
        assert oz.embedded_centers_.shape[0] == oz.cluster_centers_.shape[0]
        assert len(oz._score_clusters(self.blobs12.X)) == 9
        assert len(oz._get_cluster_sizes()) == 9

        # Image similarity
        oz.finalize()
        self.assert_images_similar(oz, tol=1.0)

    @pytest.mark.skip(reason="agglomerative not implemented yet")
    def test_ward_mds_no_legend(self):
        """
        Visual similarity with Ward, TSNE scaling, and no legend
        """
        model = AgglomerativeClustering(n_clusters=9)
        oz = InterclusterDistance(
            model, random_state=83, embedding="tsne", legend=False
        )

        # Prefit assertions
        assert_not_fitted(oz)

        assert oz.fit(self.blobs12.X) is oz  # Fit returns self

        # Postfit assertions
        assert_fitted(oz)
        assert oz.embedded_centers_.shape[0] == oz.scores_.shape[0]
        assert oz.embedded_centers_.shape[0] == oz.cluster_centers_.shape[0]
        assert len(oz._score_clusters(self.blobs12.X)) == 9
        assert len(oz._get_cluster_sizes()) == 9

        # Image similarity
        oz.finalize()
        self.assert_images_similar(oz, tol=1.0)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_quick_method(self):
        """
        Test the quick method producing a valid visualization
        """
        model = MiniBatchKMeans(3, random_state=343)
        oz = intercluster_distance(
            model, self.blobs4.X, random_state=93, legend=False, show=False
        )
        assert isinstance(oz, InterclusterDistance)

        self.assert_images_similar(oz)

    @pytest.mark.skipif(
        MPL_VERS_MAJ >= 2, reason="test requires mpl earlier than 2.0.2"
    )
    def test_legend_matplotlib_version(self, mock_toolkit):
        """
        ValueError is raised when matplotlib version is incorrect and legend=True
        """
        with pytest.raises(ImportError):
            from mpl_toolkits.axes_grid1 import inset_locator

            assert not inset_locator

        with pytest.raises(YellowbrickValueError, match="requires matplotlib 2.0.2"):
            InterclusterDistance(KMeans(), legend=True)

    @pytest.mark.skipif(
        MPL_VERS_MAJ >= 2, reason="test requires mpl earlier than 2.0.2"
    )
    def test_no_legend_matplotlib_version(self, mock_toolkit):
        """
        No error is raised when matplotlib version is incorrect and legend=False
        """
        with pytest.raises(ImportError):
            from mpl_toolkits.axes_grid1 import inset_locator

            assert not inset_locator

        InterclusterDistance(KMeans(), legend=False)

    @pytest.mark.xfail(
        reason="""third test fails with AssertionError: Expected fit
        to be called once. Called 0 times."""
    )
    def test_with_fitted(self):
        """
        Test that visualizer properly handles an already-fitted model
        """
        X, y = load_nfl(return_dataset=True).to_numpy()

        model = KMeans().fit(X, y)

        with mock.patch.object(model, "fit") as mockfit:
            oz = ICDM(model)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with mock.patch.object(model, "fit") as mockfit:
            oz = ICDM(model, is_fitted=True)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with mock.patch.object(model, "fit") as mockfit:
            oz = ICDM(model, is_fitted=False)
            oz.fit(X, y)
            mockfit.assert_called_once_with(X, y)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_within_pipeline(self):
        """
        Test that visualizer can be accessed within a sklearn pipeline
        """
        X, y = load_nfl()

        model = Pipeline(
            [
                ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
                (
                    "icdm",
                    InterclusterDistance(KMeans(5, random_state=42), random_state=42),
                ),
            ]
        )

        model.fit(X)
        model["icdm"].finalize()
        self.assert_images_similar(model["icdm"], tol=2.0)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_within_pipeline_quickmethod(self):
        """
        Test that visualizer can be accessed within a sklearn pipeline
        """
        X, y = load_nfl()

        model = Pipeline(
            [
                ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
                (
                    "icdm",
                    intercluster_distance(
                        KMeans(5, random_state=42), X, random_state=42
                    ),
                ),
            ]
        )

        model["icdm"].finalize()
        self.assert_images_similar(model["icdm"], tol=2.0)



# tests.test_datasets.test_loaders
# Test the dataset loading utilities
#
# Author:  Benjamin Bengfort <benjamin@bengfort.com>
# Created: Tue Jul 31 15:34:56 2018 -0400
#
# ID: test_loaders.py [7082742] benjamin@bengfort.com $

"""
Test the dataset loading utilities
"""

##########################################################################
## Imports
##########################################################################

import pytest
import numpy as np

from unittest.mock import patch

from yellowbrick.datasets.loaders import *
from yellowbrick.datasets.loaders import DATASETS
from yellowbrick.datasets.base import Dataset, Corpus
from yellowbrick.datasets.path import dataset_exists, dataset_archive
from yellowbrick.datasets.path import find_dataset_path
from yellowbrick.exceptions import DatasetsError

try:
    import pandas as pd
except ImportError:
    pd = None


##########################################################################
## Assertion Helpers
##########################################################################


def assert_valid_dataset(data, name):
    __tracebackhide__ = True
    assert isinstance(data, Dataset), "not a Dataset object"
    assert name in DATASETS, "dataset not in manifest"

    assert dataset_exists(name), "dataset directory does not exist"
    assert dataset_archive(
        name, DATASETS[name]["signature"]
    ), "dataset archive does not match signature"
    assert (
        find_dataset_path(name, ext=".csv.gz", raises=False) is not None
    ), "no .csv.tgz in dataset"
    assert (
        find_dataset_path(name, ext=".npz", raises=False) is not None
    ), "no .npz in dataset"

    n_files = len(data.contents())
    assert n_files == 4 or n_files == 5, "not enough files in dataset"
    assert len(data.README) > 0, "readme contains no data"
    assert len(data.meta) > 0, "metadata is empty"

    if n_files == 5:
        assert len(data.citation) > 0, "citation.bib is empty"

    assert "features" in data.meta, "no features in metadata"
    assert "target" in data.meta, "no target in metadata"


def assert_valid_corpus(corpus, name):
    __tracebackhide__ = True
    assert isinstance(corpus, Corpus), "not a Corpus object"
    assert name in DATASETS, "corpus not in manifest"

    assert dataset_exists(name), "corpus directory does not exist"
    assert dataset_archive(
        name, DATASETS[name]["signature"]
    ), "corpus archive does not match signature"

    n_contents = len(corpus.contents())
    assert n_contents > 2, "not enough files/directories in corpus"
    assert len(corpus.README) > 0, "readme contains no data"
    assert corpus.citation is None or len(corpus.citation) > 0, "citation.bib is empty"


def assert_valid_pandas(data):
    __tracebackhide__ = True
    # Get raw data frame
    df = data.to_dataframe()
    assert isinstance(df, pd.DataFrame), "raw dataframe is wrong type"

    # Get pandas data
    X, y = data.to_pandas()
    assert isinstance(X, pd.DataFrame), "X is not a DataFrame"
    assert isinstance(y, pd.Series), "y is not a Series"

    # Assert pandas is returned from to_data()
    X, y = data.to_data()
    assert isinstance(X, pd.DataFrame), "to_data does not return pandas"
    assert isinstance(y, pd.Series), "to_data does not return pandas"


def assert_valid_numpy(data):
    __tracebackhide__ = True
    X, y = data.to_numpy()
    assert isinstance(X, np.ndarray), "X is not a numpy array"
    assert isinstance(y, np.ndarray), "y is not a numpy array"
    assert X.ndim == 2 and y.ndim == 1, "X and y dimensions are incorrect"

    # Patch pandas and make defaults assertions
    X, y = data.to_data()
    assert isinstance(X, np.ndarray), "to_data does not return numpy"
    assert isinstance(y, np.ndarray), "to_data does not return numpy"

    with pytest.raises(DatasetsError):
        data.to_pandas(), "exception not raised when pandas unavailable"


##########################################################################
## Test Cases
##########################################################################


class TestDatasetLoaders(object):
    """
    Test the dataset loading functions

    Broadly: test each of the dataset loaders to ensure that they are valid
    for their particular type of dataset and that they return X and y by
    default. Then test their shape to ensure that the dataset hasn't changed
    unexpectedly between versions. See ``test_load_concrete`` for a sketch.

    Final tests with parametrize test all loaders against Base classes.
    Make sure you scroll to the bottom and implement your loader in the
    correct test batch!
    """

    def test_load_concrete(self):
        """
        Test loading the concrete regression dataset
        """
        # Load the type-specific dataset wrapper and validate it
        data = load_concrete(return_dataset=True)
        assert_valid_dataset(data, "concrete")

        # Ensure that the default returns X, y to match documentation
        # Check shape to ensure no unexpected dataset changes have occured
        # before we push something to PyPI!
        X, y = load_concrete()
        assert X.shape == (1030, 8)
        assert y.shape == (1030,)

    def test_load_energy(self):
        """
        Test loading the energy multi regression dataset
        """
        data = load_energy(return_dataset=True)
        assert_valid_dataset(data, "energy")

        X, y = load_energy()
        assert X.shape == (768, 8)
        assert y.shape == (768,)

    def test_load_credit(self):
        """
        Test loading the credit binary classification dataset
        """
        data = load_credit(return_dataset=True)
        assert_valid_dataset(data, "credit")

        X, y = load_credit()
        assert X.shape == (30000, 23)
        assert y.shape == (30000,)

    def test_load_occupancy(self):
        """
        Test loading the occupancy binary classification dataset
        """
        data = load_occupancy(return_dataset=True)
        assert_valid_dataset(data, "occupancy")

        X, y = load_occupancy()
        assert X.shape == (20560, 5)
        assert y.shape == (20560,)

    def test_load_mushroom(self):
        """
        Test loading the mushroom binary classification dataset
        """
        data = load_mushroom(return_dataset=True)
        assert_valid_dataset(data, "mushroom")

        X, y = load_mushroom()
        assert X.shape == (8123, 3)
        assert y.shape == (8123,)

    def test_load_hobbies(self):
        """
        Test loading the hobbies text corpus dataset
        """
        corpus = load_hobbies()
        assert_valid_corpus(corpus, "hobbies")

        assert len(corpus.labels) == 5
        assert len(corpus.files) == 448
        assert len(corpus.data) == 448
        assert len(corpus.target) == 448

    def test_load_game(self):
        """
        Test loading the game multiclass classification dataset
        """
        data = load_game(return_dataset=True)
        assert_valid_dataset(data, "game")

        X, y = load_game()
        assert X.shape == (67557, 42)
        assert y.shape == (67557,)

    def test_load_bikeshare(self):
        """
        Test loading the bikeshare regression dataset
        """
        data = load_bikeshare(return_dataset=True)
        assert_valid_dataset(data, "bikeshare")

        X, y = load_bikeshare()
        assert X.shape == (17379, 12)
        assert y.shape == (17379,)

    def test_load_spam(self):
        """
        Test loading the spam binary classification dataset
        """
        data = load_spam(return_dataset=True)
        assert_valid_dataset(data, "spam")

        X, y = load_spam()
        assert X.shape == (4600, 57)
        assert y.shape == (4600,)

    def test_load_walking(self):
        """
        Test loading the walking activity clustering dataset
        """
        data = load_walking(return_dataset=True)
        assert_valid_dataset(data, "walking")

        X, y = load_walking()
        assert X.shape == (149332, 4)
        assert y.shape == (149332,)

    def test_load_nfl(self):
        """
        Test loading the nfl clustering dataset
        """
        data = load_nfl(return_dataset=True)
        assert_valid_dataset(data, "nfl")

        X, y = load_nfl()
        assert X.shape == (494, 23)
        assert y.shape == (494,)

    @pytest.mark.skipif(pd is None, reason="pandas is required for this test")
    @pytest.mark.parametrize(
        "loader",
        [
            load_bikeshare,
            load_concrete,
            load_credit,
            load_energy,
            load_game,
            load_mushroom,
            load_occupancy,
            load_spam,
            load_walking,
            load_nfl,
        ],
        ids=lambda l: l.__name__,
    )
    def test_load_pandas(self, loader):
        """
        Test loading datasets as pandas objects
        """
        data = loader(return_dataset=True)
        assert_valid_pandas(data)

    @patch("yellowbrick.datasets.base.pd", None)
    @pytest.mark.parametrize(
        "loader",
        [
            load_bikeshare,
            load_concrete,
            load_credit,
            load_energy,
            load_game,
            load_mushroom,
            load_occupancy,
            load_spam,
            load_walking,
            load_nfl,
        ],
        ids=lambda l: l.__name__,
    )
    def test_load_numpy(self, loader):
        """
        Test loading datasets as numpy defaults
        """
        data = loader(return_dataset=True)
        assert_valid_numpy(data)



# tests.test_pipeline
# Tests to ensure that the visual pipeline works as expected.
#
# Author:   Benjamin Bengfort <bbengfort@districtdatalabs.com>
# Created:  Fri Oct 07 22:10:50 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_pipeline.py [1efae1f] benjamin@bengfort.com $

"""
Tests to ensure that the visual pipeline works as expected.
"""

##########################################################################
## Imports
##########################################################################

import os
import pytest
import sklearn

from unittest import mock
from yellowbrick.base import Visualizer
from yellowbrick.pipeline import VisualPipeline
from sklearn.pipeline import Pipeline
from sklearn.base import BaseEstimator, TransformerMixin


##########################################################################
## Mock Objects
##########################################################################


class Thing(object):
    pass


class MockEstimator(BaseEstimator):
    def fit(self, X, y=None, **kwargs):
        return self


class MockVisualEstimator(Visualizer):
    def fit(self, X, y=None, **kwargs):
        self.draw(**kwargs)
        return self

    def draw(self, **kwargs):
        pass


class MockTransformer(BaseEstimator, TransformerMixin):
    def fit(self, X, y=None, **kwargs):
        return self

    def transform(self, X, **kwargs):
        return X


class MockVisualTransformer(Visualizer, TransformerMixin):
    def fit(self, X, y=None, **kwargs):
        self.draw(**kwargs)
        return self

    def transform(self, X, **kwargs):
        return X

    def draw(self, **kwargs):
        pass


##########################################################################
## VisualPipeline Tests
##########################################################################


class TestVisualPipeline(object):
    def test_validate_steps(self):
        """
        Assert that visual transformers can be added to pipelines
        """

        # Pipeline objects have a _validate_steps method that raises an
        # TypeError if the steps don't match transforms --> estimator.
        # NOTE: sklearn 1.7+ removed eager validation at construction time

        sklearn_version = tuple(map(int, sklearn.__version__.split(".")[:2]))

        # validate a bad intermediate transformer on the Pipeline
        if sklearn_version < (1, 7):
            with pytest.raises(TypeError):
                Pipeline(
                    [
                        ("real", MockTransformer()),
                        ("bad", Thing()),
                        ("model", MockEstimator()),
                    ]
                )

        # validate a bad intermediate transformer on the VisualPipeline
        if sklearn_version < (1, 7):
            with pytest.raises(TypeError):
                VisualPipeline(
                    [
                        ("real", MockTransformer()),
                        ("bad", Thing()),
                        ("model", MockEstimator()),
                    ]
                )

        # validate a bad final estimator on the Pipeline
        if sklearn_version < (1, 7):
            with pytest.raises(TypeError):
                Pipeline([("real", MockTransformer()), ("bad", Thing())])

        # validate a bad final estimator on the VisualPipeline
        if sklearn_version < (1, 7):
            with pytest.raises(TypeError):
                VisualPipeline([("real", MockTransformer()), ("bad", Thing())])

        # validate visual transformers on a Pipeline
        try:
            Pipeline(
                [
                    ("real", MockTransformer()),
                    ("visual", MockVisualTransformer()),
                    ("model", MockEstimator()),
                ]
            )
        except TypeError:
            self.fail("could not add a visual transformer to a Pipeline!")

        # validate visual transformers on a VisualPipeline
        try:
            VisualPipeline(
                [
                    ("real", MockTransformer()),
                    ("visual", MockVisualTransformer()),
                    ("model", MockEstimator()),
                ]
            )
        except TypeError:
            self.fail("could not add a visual transformer to a VisualPipeline!")

    def test_visual_steps_property(self):
        """
        Test the visual steps property to filter visualizers
        """

        pipeline = VisualPipeline(
            [
                ("a", MockTransformer()),
                ("b", MockVisualTransformer()),
                ("c", MockTransformer()),
                ("d", MockVisualTransformer()),
                ("e", MockEstimator()),
            ]
        )

        assert "a" not in pipeline.visual_steps
        assert "b" in pipeline.visual_steps
        assert "c" not in pipeline.visual_steps
        assert "d" in pipeline.visual_steps
        assert "e" not in pipeline.visual_steps

    def test_pipeline_show(self):
        """
        Test the show call against the VisualPipeline
        """

        pipeline = VisualPipeline(
            [
                ("a", mock.MagicMock(MockTransformer())),
                ("b", mock.MagicMock(MockVisualTransformer())),
                ("c", mock.MagicMock(MockTransformer())),
                ("d", mock.MagicMock(MockVisualTransformer())),
                ("e", mock.MagicMock(MockEstimator())),
            ]
        )

        pipeline.show()
        pipeline.steps[1][1].show.assert_called_once_with(outpath=None)
        pipeline.steps[3][1].show.assert_called_once_with(outpath=None)

    def test_pipeline_savefig_show(self):
        """
        Test the show call with an outdir to save all the figures
        """
        pipeline = VisualPipeline(
            [
                ("a", mock.MagicMock(MockTransformer())),
                ("b", mock.MagicMock(MockVisualTransformer())),
                ("c", mock.MagicMock(MockTransformer())),
                ("d", mock.MagicMock(MockVisualTransformer())),
                ("e", mock.MagicMock(MockVisualEstimator())),
            ]
        )

        # Must use path joining for Windows compatibility
        tmpdir = os.path.join("tmp", "figures")

        pipeline.show(outdir=tmpdir)
        pipeline.steps[1][1].show.assert_called_once_with(
            outpath=os.path.join(tmpdir, "b.pdf")
        )
        pipeline.steps[3][1].show.assert_called_once_with(
            outpath=os.path.join(tmpdir, "d.pdf")
        )
        pipeline.steps[4][1].show.assert_called_once_with(
            outpath=os.path.join(tmpdir, "e.pdf")
        )

    @pytest.mark.skip(reason="need to find a way for fit to return self in mocks")
    def test_fit_transform_show_and_draw_calls(self):
        """
        Test calling fit, transform, and show on the pipeline
        """

        pipeline = VisualPipeline(
            [
                ("a", mock.MagicMock(MockTransformer())),
                ("b", mock.MagicMock(MockVisualTransformer())),
                ("c", mock.MagicMock(MockTransformer())),
                ("d", mock.MagicMock(MockVisualTransformer())),
                ("e", mock.MagicMock(MockEstimator())),
            ]
        )

        X = [[1, 1, 1, 1, 1], [2, 2, 2, 2, 2], [3, 3, 3, 3, 3]]

        y = [1, 2, 3, 4, 5]

        pipeline.fit(X, y)
        for name, step in pipeline.named_steps.items():
            step.fit.assert_called_once_with(X, y)

        pipeline.transform(X)
        for name, step in pipeline.named_steps.items():
            if name == "e":
                continue
            step.transform.assert_called_once_with(X)

        pipeline.show()
        for name, step in pipeline.named_steps.items():
            if name in {"a", "c", "e"}:
                continue
            step.show.assert_called_once_with(outpath=None)



# tests.test_regressor.test_residuals
# Ensure that the regressor residuals visualizations work.
#
# Author:   Rebecca Bilbro
# Author:   Benjamin Bengfort
# Created:  Sat Oct 8 16:30:39 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_residuals.py [7d3f5e6] benjamin@bengfort.com $

"""
Ensure that the regressor residuals visualizations work.
"""

##########################################################################
## Imports
##########################################################################

import sys
import pytest
import matplotlib as mpl
import matplotlib.pyplot as plt

from yellowbrick.datasets import load_energy
from yellowbrick.exceptions import YellowbrickValueError
from yellowbrick.regressor.residuals import ResidualsPlot, residuals_plot

from unittest import mock
from tests.fixtures import Dataset, Split
from tests.base import IS_WINDOWS_OR_CONDA, VisualTestCase

from sklearn.datasets import make_regression
from sklearn.linear_model import Ridge, Lasso
from sklearn.neural_network import MLPRegressor
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import train_test_split as tts
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler


try:
    import pandas as pd
except ImportError:
    pd = None

# Determine version of matplotlib
MPL_VERS_MAJ = int(mpl.__version__.split(".")[0])


##########################################################################
## Data
##########################################################################


@pytest.fixture(scope="class")
def data(request):
    """
    Creates a fixture of train and test splits for the sklearn digits dataset
    For ease of use returns a Dataset named tuple composed of two Split tuples.
    """
    X, y = make_regression(
        n_samples=500,
        n_features=22,
        n_informative=8,
        random_state=42,
        noise=0.2,
        bias=0.2,
    )

    X_train, X_test, y_train, y_test = tts(X, y, test_size=0.2, random_state=11)

    # Set a class attribute for digits
    request.cls.data = Dataset(Split(X_train, X_test), Split(y_train, y_test))

##########################################################################
## Residuals Plot Test Cases
##########################################################################


@pytest.mark.usefixtures("data")
class TestResidualsPlot(VisualTestCase):
    """
    Test ResidualPlot visualizer
    """

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_residuals_plot(self):
        """
        Image similarity of residuals plot on random data with OLS
        """
        _, ax = plt.subplots()

        visualizer = ResidualsPlot(LinearRegression(), ax=ax)

        visualizer.fit(self.data.X.train, self.data.y.train)
        visualizer.score(self.data.X.test, self.data.y.test)
        visualizer.finalize()

        self.assert_images_similar(visualizer)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_residuals_plot_QQ_plot(self):
        """
        Image similarity of residuals and Q-Q plot on random data with OLS
        """
        _, ax = plt.subplots()

        visualizer = ResidualsPlot(LinearRegression(), hist=False, qqplot=True, ax=ax)

        visualizer.fit(self.data.X.train, self.data.y.train)
        visualizer.score(self.data.X.test, self.data.y.test)

        self.assert_images_similar(visualizer)

    def test_either_hist_or_QQ_plot(self):
        """
        Setting both hist=True and qqplot=True raises exception.
        """
        with pytest.raises(
            YellowbrickValueError, match="Set either hist or qqplot to False"
        ):
            ResidualsPlot(LinearRegression(), hist=True, qqplot=True)

    @pytest.mark.xfail(
        sys.platform == "win32", reason="images not close on windows (RMSE=32)"
    )
    @pytest.mark.filterwarnings("ignore:Stochastic Optimizer")
    def test_residuals_plot_no_histogram(self):
        """
        Image similarity test when hist=False
        """
        _, ax = plt.subplots()

        model = MLPRegressor(random_state=19)
        visualizer = ResidualsPlot(model, ax=ax, hist=False)

        visualizer.fit(self.data.X.train, self.data.y.train)
        visualizer.score(self.data.X.test, self.data.y.test)
        visualizer.finalize()

        self.assert_images_similar(visualizer, tol=1, remove_legend=True)

    @pytest.mark.skipif(
        MPL_VERS_MAJ >= 2, reason="test requires mpl earlier than 2.0.2"
    )
    def test_hist_matplotlib_version(self, mock_toolkit):
        """
        ValueError is raised when matplotlib version is incorrect and hist=True
        """
        with pytest.raises(ImportError):
            from mpl_toolkits.axes_grid1 import make_axes_locatable

            assert not make_axes_locatable

        with pytest.raises(YellowbrickValueError, match="requires matplotlib 2.0.2"):
            ResidualsPlot(LinearRegression(), hist=True)

    @pytest.mark.skipif(
        MPL_VERS_MAJ >= 2, reason="test requires mpl earlier than 2.0.2"
    )
    def test_no_hist_matplotlib_version(self, mock_toolkit):
        """
        No error is raised when matplotlib version is incorrect and hist=False
        """
        with pytest.raises(ImportError):
            from mpl_toolkits.axes_grid1 import make_axes_locatable

            assert not make_axes_locatable

        ResidualsPlot(LinearRegression(), hist=False)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_residuals_quick_method(self):
        """
        Image similarity test using the residuals plot quick method
        """
        _, ax = plt.subplots()

        model = Lasso(random_state=19)
        oz = residuals_plot(
            model,
            self.data.X.train,
            self.data.y.train,
            self.data.X.test,
            self.data.y.test,
            ax=ax,
            line_color="#cccccc",
            test_color="r",
            train_color="y",
            show=False,
        )

        assert isinstance(oz, ResidualsPlot)
        self.assert_images_similar(oz)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_residuals_quick_method_train_only(self):
        """
        Test the quick method with only train data (simplest args)
        """
        oz = residuals_plot(
            Ridge(random_state=19), self.data.X.train, self.data.y.train, show=False
        )

        assert isinstance(oz, ResidualsPlot)
        self.assert_images_similar(oz)

    def test_residuals_quick_method_missing_data(self):
        """
        Ensure the quick method requires both X_test and y_test if one specified
        """
        msg = "both X_test and y_test are required if one is specified"
        with pytest.raises(YellowbrickValueError, match=msg):
            residuals_plot(
                Lasso(),
                self.data.X.train, self.data.y.train,
                self.data.X.test, show=False
            )

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    @pytest.mark.skipif(pd is None, reason="pandas is required")
    def test_residuals_plot_pandas(self):
        """
        Test Pandas real world dataset with image similarity on Lasso
        """
        _, ax = plt.subplots()

        # Load the occupancy dataset from fixtures
        data = load_energy(return_dataset=True)
        X, y = data.to_pandas()

        # Create train/test splits
        splits = tts(X, y, test_size=0.2, random_state=231)
        X_train, X_test, y_train, y_test = splits

        visualizer = ResidualsPlot(Lasso(random_state=44), ax=ax)
        visualizer.fit(X_train, y_train)
        visualizer.score(X_test, y_test)
        visualizer.finalize()

        self.assert_images_similar(visualizer)

    def test_residuals_plot_numpy(self):
        """
        Test NumPy real world dataset with image similarity on Lasso
        """
        _, ax = plt.subplots()

        # Load the occupancy dataset from fixtures
        data = load_energy(return_dataset=True)
        X, y = data.to_numpy()

        # Create train/test splits
        splits = tts(X, y, test_size=0.2, random_state=231)
        X_train, X_test, y_train, y_test = splits

        visualizer = ResidualsPlot(Lasso(random_state=44), ax=ax)
        visualizer.fit(X_train, y_train)
        visualizer.score(X_test, y_test)
        visualizer.finalize()

        self.assert_images_similar(visualizer, tol=1.5)

    def test_score(self):
        """
        Assert returns R2 score
        """
        visualizer = ResidualsPlot(Ridge(random_state=8893))

        visualizer.fit(self.data.X.train, self.data.y.train)
        score = visualizer.score(self.data.X.test, self.data.y.test)

        assert score == pytest.approx(0.9999888484, rel=1e-4)
        assert visualizer.train_score_ == pytest.approx(0.9999906, rel=1e-4)
        assert visualizer.test_score_ == score

    @mock.patch("yellowbrick.regressor.residuals.plt.sca", autospec=True)
    def test_alpha_param(self, mock_sca):
        """
        Test that the user can supply an alpha param on instantiation
        """
        # Instantiate a prediction error plot, provide custom alpha
        visualizer = ResidualsPlot(
            Ridge(random_state=8893), train_alpha=0.3, test_alpha=0.75, hist=False
        )
        alphas = {"train_point": 0.3, "test_point": 0.75}
        # Test param gets set correctly
        assert visualizer.alphas == alphas

        visualizer.ax = mock.MagicMock()
        visualizer.fit(self.data.X.train, self.data.y.train)
        visualizer.score(self.data.X.test, self.data.y.test)

        # Test that alpha was passed to internal matplotlib scatterplot
        _, scatter_kwargs = visualizer.ax.scatter.call_args
        assert "alpha" in scatter_kwargs
        assert scatter_kwargs["alpha"] == 0.75

    def test_is_fitted_param(self):
        """
        Test that the user can supply an is_fitted param and it's state is maintained
        """
        # Instantiate a prediction error plot, provide custom is_fitted
        visualizer = ResidualsPlot(Ridge(random_state=8893), is_fitted=False)

        # Test param gets set correctly
        assert visualizer.is_fitted == False
    
    @pytest.mark.xfail(
        reason="""third test fails with AssertionError: Expected fit
        to be called once. Called 0 times."""
    )
    def test_residuals_with_fitted(self):
        """
        Test that ResidualsPlot properly handles an already-fitted model
        """
        X, y = load_energy(return_dataset=True).to_numpy()

        model = Ridge().fit(X, y)

        with mock.patch.object(model, "fit") as mockfit:
            oz = ResidualsPlot(model)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with mock.patch.object(model, "fit") as mockfit:
            oz = ResidualsPlot(model, is_fitted=True)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with mock.patch.object(model, "fit") as mockfit:
            oz = ResidualsPlot(model, is_fitted=False)
            oz.fit(X, y)
            mockfit.assert_called_once_with(X, y)

    def test_within_pipeline(self):
        """
        Test that visualizer can be accessed within a sklearn pipeline
        """
        X, y = load_energy(return_dataset=True).to_numpy()
        X_train, X_test, y_train, y_test = tts(X, y, test_size=0.2, random_state=42)

        model = Pipeline([
            ('minmax', MinMaxScaler()),
            ('rp', ResidualsPlot(Ridge(random_state=8893)))
        ])

        model.fit(X_train, y_train)
        model.score(X_test, y_test)
        model['rp'].finalize()
        self.assert_images_similar(model['rp'], tol=3.0)

    def test_within_pipeline_quickmethod(self):
        """
        Test that visualizer quickmethod can be accessed within a
        sklearn pipeline
        """
        X, y = load_energy(return_dataset=True).to_numpy()
        X_train, X_test, y_train, y_test = tts(X, y, test_size=0.2, random_state=42)

        model = Pipeline([
            ('minmax', MinMaxScaler()),
            ('rp', residuals_plot(Ridge(random_state=8893), X_train, y_train, X_test, y_test))
        ])

        model.fit(X_train, y_train)
        model.score(X_test, y_test)
        model['rp'].finalize()
        self.assert_images_similar(model['rp'], tol=3.0)

    def test_pipeline_as_model_input(self):
        """
        Test that visualizer can handle sklearn pipeline as model input
        """
        X, y = load_energy(return_dataset=True).to_numpy()
        X_train, X_test, y_train, y_test = tts(X, y, test_size=0.2, random_state=42)

        model = Pipeline([
            ('minmax', MinMaxScaler()),
            ('ridge', Ridge(random_state=8893))
        ])

        oz = ResidualsPlot(model)
        oz.fit(X_train, y_train)
        oz.score(X_test, y_test)
        oz.finalize()
        self.assert_images_similar(oz, tol=3.0)

    def test_pipeline_as_model_input_quickmethod(self):
        """
        Test that visualizer can handle sklearn pipeline as model input
        within a quickmethod
        """
        X, y = load_energy(return_dataset=True).to_numpy()
        X_train, X_test, y_train, y_test = tts(X, y, test_size=0.2, random_state=42)

        model = Pipeline([
            ('minmax', MinMaxScaler()),
            ('ridge', Ridge(random_state=8893))
        ])

        oz = residuals_plot(model, X_train, y_train, X_test, y_test, show=False)
        self.assert_images_similar(oz, tol=3.0)



    
